/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;

import fusion.manager.aggregation.UnimodalPotentialFunction;

public class Feature implements FeatureServerInterface  {

    //socket server port on which it will listen
    private int portIn;
    private int portOut;
    
    private FeatureInThread inThread;
    private FeatureOutThread outThread;
    
    private boolean isTrained = false;
    
    private Object featureValue = null;
    public Double memValue = null;
	
	String UID;
	List<Sensor> sensors;
	
	Algorithm algorithm;
	private Class<?> AlgClass;
	private Object AlgClassObject;
	private Method m;
	
	private List<Double> trainData;
	
	UnimodalPotentialFunction memFunc;
	
	public Feature(String UID, Algorithm algorithm, int port){

		//Semantics
		this.UID = UID;
		this.algorithm = algorithm;
		this.portIn = port;
		this.portOut = port + 1;
		loadAlgorithmClass();
		
		//Fusion
		trainData = new ArrayList<Double>();
		sensors = new ArrayList<Sensor>();		
		memFunc = new UnimodalPotentialFunction();
	}
	
	//################################################################################################
	/**
	 * Get the UID of this feature
	 */
	public String getUID(){
		return UID;
	}
	
	/**
	 * Returns the list of solid sensors that are incorporated for this feature
	 */
	public List<Sensor> getSensors(){
		return sensors;
	}
	
	/**
	 * Add sensor objects to feature
	 */
	public void addSensor(Sensor sensor){
		sensors.add(sensor);
	}

	/**
	 * Returns the algorithm object that computes the value of this feature
	 */
	@Override
	public Algorithm getAlgorithm() {
		return algorithm;
	}
	
	@Override
	public Object computeFeatureValue(Object data){
		
		// Feature computation
		featureValue = computeValue(data);
		//System.out.println("Algorithm returns: " + featureValue);
		
		if(featureValue!=null && !isTrained){
			trainData.add((Double) featureValue);
			//Train a membership function 
			if(trainData.size() == 100){
				trainMembershipFunction(trainData);
			}
		}
		else if(isTrained){
			memValue = computeMembershipValue((double) featureValue);
			//System.out.println("Memebership: " + memValue);
		}
		
		return featureValue;
	}
	
	/**
	 * Initialize Inputs and Outputs of the feature. 
	 */
	public void initIOThreads(){
		inThread = new FeatureInThread(portIn, this);
		inThread.start();
		
		outThread = new FeatureOutThread(portOut, this);
		outThread.start();
	}
	
	/**
	 * Close I/O threads of a feature
	 */
	public void closeThreads(){
		inThread.closeThread();
		outThread.closeThread();
	}

	/**
	 * Train the membership function of this feature. Parameters are pCe: elementary fuzziness (expert knowledge)
	 * and D: steepness (expert knowledge) are fixed (automatic parameterization should be implemented).
	 */
	@Override
	public void trainMembershipFunction(List<Double> data) {
		// TODO Auto-generated method stub
		memFunc.train(data);
		isTrained=true;
	}
	
	/**
	 * Compute the membership of the given value.
	 * @param value: value for which the membership value is computed
	 * @return
	 */
	@Override
	public double computeMembershipValue(double input){
		return memFunc.computeMembershipValue(input);
	}

	@Override
	public boolean isTrained() {
		return isTrained;
	}

	/**
	 * Load the algorithm object from the path of the compiled class file
	 * 
	 */
	@Override
	public void loadAlgorithmClass(){
		String dir = System.getProperty("user.dir");
		//Load Class
    	try {
    		//Path to java file
	    	File file = new File(dir + "/ExternalJavaFiles" + java.io.File.separator);
	    	// Convert file to an URL
		    URI uri = file.toURI();
		    URL url = uri.toURL();
		    URL[] urls = new URL[]{url};
		    		    
		    //initialise class loader
			@SuppressWarnings("resource")
			ClassLoader classLoader = new URLClassLoader(urls);
			
			//load algorithm class
			System.out.println("Load algorithms." + algorithm.getUID());
            AlgClass = classLoader.loadClass("algorithms." + algorithm.getUID());
            
            //Call constructor and get run-method
            Constructor<?> constructor = AlgClass.getConstructor();
            AlgClassObject = constructor.newInstance();
	        m = AlgClass.getMethod("run", double.class);
        } catch (Exception e) {
            e.printStackTrace();
        }
	}
	
	/**
	 * Takes input and computes the feature value
	 * @param input
	 * @return
	 */
	@Override
	public Object computeValue(Object input){
		try {
			double in = (double) input;
	        Object result = m.invoke(AlgClassObject, in);
	        //System.out.println("Result: " + result);
	        return result;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
}
