/*
 * Copyright 2017 inIT - Institute Industrial IT

This file is part of "Automated Fusion System Design and Adaptation Implementation.

"Automated Fusion System Design and Adaptation Implementation is free software: 
you can redistribute it and/or modify it under the terms of the GNU General Public 
License as published by the Free Software Foundation, either version 3 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this 
program.  If not, see <http://www.gnu.org/licenses/>.

* A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
*/
package fusion.manager;

import java.util.ArrayList;
import java.util.List;

public class FeatureStore {
	//List of features
	private List<Feature> features;
	FeatureConfigThread confThread;
	List<Algorithm> algorithms;
	
	/**
	 * The FeatureStore holds available features including associated sensors and algorithms.
	 */
	public FeatureStore(List<Algorithm> algorithms){		
		features = new ArrayList<Feature>();
		this.algorithms = algorithms;
	}		
	
	/**
	 * Returns a list of available features
	 * @return
	 */
	public List<Feature> getFeatureList(){
		return features;
	}
	
	/**
	 * Returns the feature with the specific UID
	 * @param UID
	 * @return
	 */
	public Feature getFeature(String UID){
		for(int i=0; i<features.size(); i++){
			if(features.get(i).getUID().equals(UID)){
				return features.get(i);
			}
		}		
		return null;
	}
	
	/**
	 * Returns the feature with specific index	
	 * @param index
	 * @return
	 */
	public Feature getFeature(int index){
		return features.get(index);
	}
	
	/**
	 * Returns a list of features that incorporate the sensor with a specific uid.
	 * @param uid
	 * @return
	 */
	public List<Feature> getFeaturesWithSensor(String uid){
		List<Feature> featuresSens  = new ArrayList<Feature>();
		
		for(int i=0; i<features.size();i++){
			for(int j=0; j<features.get(i).getSensors().size(); j++){
				String uidS = features.get(i).getSensors().get(j).getUID();
				if(uidS.equals(uid)){
					featuresSens.add(features.get(i));
				}
			}
		}		
		return featuresSens;
	}
	
	/**
	 * Returns a list of features that use the algorithm with a specific uid.
	 * @param uid
	 * @return
	 */
	public List<Feature> getFeaturesWithAlgorithm(String uid){
		List<Feature> featuresAlg  = new ArrayList<Feature>();
		
		for(int i=0; i<features.size();i++){
			String uidA = features.get(i).getAlgorithm().getUID();
			if(uidA.equals(uid)){
				featuresAlg.add(features.get(i));
			}
		}		
		return featuresAlg;
	}
	
	/**
	 * Returns the algorithm with the particular uid.
	 * @param uid
	 * @return
	 */
	public Algorithm getAlgorithm(String uid){
		
		for(int i=0; i<algorithms.size();i++){
			String uidA = algorithms.get(i).getUID();
			if(uidA.equals(uid)){
				return algorithms.get(i);
			}
		}		
		return null;
	}
	
	/**
	 * Add a feature to the feature store.
	 * @param feature
	 */
	public void addFeature(Feature feature){
		feature.initIOThreads();
		features.add(feature);		
	}
	
	/**
	 * Check if feature with specific UID already exists
	 * @return
	 */
	public boolean isFeatureExisting(String UID){
		boolean isExisting = false;
		
		for(int i = 0; i<features.size(); i++){
			if(features.get(i).getUID().equals(UID))
				isExisting=true;
		}		
		return isExisting;
	}
	
	/**
	 * Clears the list of features from the FeatureStore
	 */
	public void clear(){
		features.clear();
	}
	
	/**
	 * Starts a thread, which waits for incoming messages for feature configuration.
	 */
	public void initConfigThread(){
		confThread = new FeatureConfigThread(9900, this);
		confThread.start();
	}
	
	/**
	 * Close the threads for feature configuration and feature I/Os.
	 */
	public void closeConfigThread(){
		for(int i=0; i<features.size();i++){
			features.get(i).closeThreads();
		}
		confThread.closeThread();
	}
	
	
}
