/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager;

import gpioreader.Pi4JUSBSer;

public class Sensor implements SensorInterface {
	
	// check if data interface is initialized
	boolean isInitDataInterface = false;
	
	private String UID;
	private String physPheno;
	private int dimensionality;
	private String foi;
	private String dataType;
	private int dataLength;
	private String sourceIP;
	private int port;
	private String interfaceType;
	private int interfaceChannel;
	private SensorThread sensThread = null;
	
	/**
	 * The Sensor-Class represents an elementary sensor of an intelligent sensor. Semantic data is required to generate an sensor object.
	 * For data exchange an interface can be created using the initDataInterface-Class.
	 * This is not required if only semantics are required.
	 * 
	 * @param UID Distinct name of the elementary sensor
	 * @param suitablePhysPheno Physical phenomenon the sensor states about
	 * @param signalDimensionality Dimensionality of the output signal
	 * @param dataType Data type of the returned measurement
	 * @param sourceIP Network address (same as for the intelligent sensor)
	 * @param observedEntity The entity (Module, area, product, etc.) the sensor observes
	 */
	public Sensor(String UID, String suitablePhysPheno, int signalDimensionality, String dataType, 
			int dataLength, String sourceIP, String interfaceType, int interfaceChannel,
			int port, String observedEntity){

		//Semantics
		this.UID = UID;
		this.physPheno = suitablePhysPheno;
		this.foi = observedEntity;
		this.dimensionality = signalDimensionality;
		this.dataType = dataType;
		this.dataLength = dataLength;
		this.sourceIP = sourceIP;
		this.port = port;
		this.interfaceType = interfaceType;
		this.interfaceChannel = interfaceChannel;
		
	}
	
	@Override
	public String getUID() {
		return UID;
	}

	@Override
	public String getfoi() {		
		return foi;
	}

	@Override
	public String getPhysPheno() {
		return physPheno;
	}

	@Override
	public int getOutputDimensonality() {
		return dimensionality;
	}	

	@Override
	public String getDataType() {
		return dataType;
	}
	
	@Override
	public int getDataLength() {
		return dataLength;
	}

	@Override
	public String getIP() {
		return sourceIP;
	}
	

	/**
	 * Initialize the interface (GPIO, USB, etc.) for sensor data propagation. Only required for intelligent sensor nodes.
	 * If it is not called only semantic information are available from the sensor object.
	 */
	@Override
	public void initDataInterface() {		
		isInitDataInterface = true;
				
		// Initialize thread for sensor data
		if(interfaceType.equals("GPIO")){
			sensThread = new SensorThread(1, interfaceChannel, port);
			sensThread.start();
		}
		else if(interfaceType.equals("UART")){
			sensThread = new SensorThread(2, interfaceChannel, port);
			sensThread.start();
		}		
	}
	
	public void initUSBDataInterface(Pi4JUSBSer usbSerial) {
		
		sensThread = new SensorThread(3, interfaceChannel, port);
		sensThread.setUSBSerial(usbSerial);		
		sensThread.start();
		
	}
	
	public void closeSensorInterface(){
		sensThread.interrupt();
		sensThread.closeThread();
	}

	/**
	 * Returns the actual sensor measurement of generic data type.
	 */
	@Override
	public Object getSensorData() {
		return sensThread.getCurrentSensValue();
	}

	@Override
	public int getPort() {
		return port;
	}

	@Override
	public String getInterfaceType() {
		// TODO Auto-generated method stub
		return interfaceType;
	}

	@Override
	public int getInterfaceChannel() {
		// TODO Auto-generated method stub
		return interfaceChannel;
	}

	
}
