/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */

/**
 * This class represents a unimodal potential function.
 */

package fusion.manager.aggregation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;


/**
 * 
 * @author Alexander Fritze
 */

public class UnimodalPotentialFunction extends FuzzyMembershipFunction
{  
	
	private double pCe;											// elementary fuzziness (expert knowledge)
	private double  D;											// steepness (expert knowledge)
	private List<Double> trainData = new ArrayList<Double>();		// measurements for training
	private List<Double> parameters = new ArrayList<Double>();
	
	private double currentMembershipValue;
	
    public UnimodalPotentialFunction()
    {
        super();
    }
    
	
	/**
	 *  This function returns the parameters for membership function of one sensor
	 * @param senData : Trainig data
	 * @param pCe : elementary fuzziness (expert knowledge)
	 * @param D : steepness (expert knowledge)
	 * @return
	 */
	public void train(List<Double> trainData){
		parameters.clear();
		this.D = 2;
		this.pCe = 1;	
		this.trainData =trainData;
		this.currentMembershipValue = 0;
		
		double S = computeS();
		double C = computeC();
		
		parameters.add(S);
		parameters.add(C);
		parameters.add(this.D);
		parameters.add(this.pCe);
		
		this.setTrained(true);
	}	
	
	/**
	 *  Compute the average value S
	 * @return
	 */
	private double computeS(){
		double S = 0;
		double sum = 0;
		int N = trainData.size();
		
		for(Double val : trainData)
		{
			sum = sum + val;
		}
		S = sum/N;		
		return S;
	}
	
	/**
	 *  Compute the width C
	 * @return
	 */
	private double computeC(){
		double C = 0;
		double max = Collections.max(trainData);
		double min = Collections.min(trainData);
		
		C = (double) ((1 + (2 * pCe)) * 0.5 * (max - min));
		
		return C;
	}
	
	/**
	 * Compute the basic belief assignment of the feature value
	 * @param value : feature value
	 * @return
	 */	 
	public double computeMembershipValue(double value){
		//System.out.println("Training: " +trainData + "\n Para: " + parameters);
		if(this.isTrained()){
			double m = 0;
			//Log.d("BBAs", "Params: " + parameters.toString());
			
			double a = Math.abs(value - parameters.get(0));
			//Log.d("BBAs","a: " + Double.toString(a));
			
			double b;
			double d;
			if(parameters.get(1).equals(Double.valueOf(0.0))){
				//System.out.println("XXX");
				 b =  0;
				 d = 0;
			}
			else{
				b = a / parameters.get(1);
				d = (double) Math.pow( b, parameters.get(2));
			}
			
			//Log.d("BBAs","b: " + Double.toString(b));	
			//Log.d("BBAs","d: " + Double.toString(d));
			
			double m1 = (double)(Math.pow(2, (d)));
			//Log.d("BBAs","2^d: " + Double.toString(m1));
			
			//System.out.println("Input: " + value + " a:" + a + " b:" + b + " d: " + d + " m1:"
					//+ m1);
			
			m = 1 / m1;
			currentMembershipValue = m;
			return m;
		}
		else{
			System.out.println("Membership funtion is not trained. Cannot return membership value.");
			return -1;
		}
		
	}
	
	public double getCurrentMembershipValue(){
		return currentMembershipValue;
	}
}