/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package opcua.server;

import java.io.FileNotFoundException;

import org.opcfoundation.ua.builtintypes.LocalizedText;
import org.opcfoundation.ua.builtintypes.NodeId;
import org.opcfoundation.ua.builtintypes.QualifiedName;
import org.opcfoundation.ua.builtintypes.UnsignedInteger;
import org.opcfoundation.ua.core.Identifiers;
import org.vast.xml.XMLReaderException;
import com.prosysopc.ua.StatusException;
import com.prosysopc.ua.nodes.UaNodeFactoryException;
import com.prosysopc.ua.nodes.UaObject;
import com.prosysopc.ua.samples.server.MyEventManagerListener;
import com.prosysopc.ua.server.NodeManagerUaNode;
import com.prosysopc.ua.server.UaInstantiationException;
import com.prosysopc.ua.server.UaServer;
import com.prosysopc.ua.types.opcua.server.FolderTypeNode;

import fileManager.Read;
import opcua.types.Algorithm_Type;
import opcua.types.IntelligentSensor_Type;
import opcua.types.Sensor_Type;

/**
 * A sample customized node manager, which actually just overrides the standard
 * NodeManagerUaNode and initializes the nodes for the demo.
 */
public class IntelligentSensorNodeManager extends NodeManagerUaNode {
	public static final String NAMESPACE = "http://www.init-owl.de/IntelligentSensor/";
	private FolderTypeNode myObjectsFolder;
	final MyEventManagerListener myEventManagerListener = new MyEventManagerListener();
	
	Read reader;

	/**
	 * Create instance of IntelligentSensorNodeManager
	 *
	 * @param server
	 *            the server in which the node manager is created.
	 * @param namespaceUri
	 *            the namespace URI for the nodes
	 * @throws Exception
	 */
	public IntelligentSensorNodeManager(UaServer server, String namespaceUri, Read reader)
			throws Exception {
		super(server, namespaceUri);
		this.reader = reader;
	}

	private void createAddressSpace() throws Exception {
		// +++ My nodes +++		
		int ns = getNamespaceIndex();
		//System.out.println(ns);
		//int ns = server.getNamespaceTable().getIndex("http:/init-owl.de/IntelligentSensor/");

		// UA types and folders which we will use
		final UaObject objectsFolder = getServer().getNodeManagerRoot()
				.getObjectsFolder();

		// Folder for my objects
		final NodeId myObjectsFolderId = new NodeId(ns, "MyObjectsFolder");
		myObjectsFolder = createInstance(FolderTypeNode.class, "MyObjects", myObjectsFolderId);
		
		this.addNodeAndReference(objectsFolder, myObjectsFolder, Identifiers.Organizes);
		
		try {
			
			//Read reader = new Read(folderSelfDescription);
			
			//IntelligentSensor_TypeNode test = createInstance(IntelligentSensor_TypeNode.class, reader.getNameIsens());
			
			IntelligentSensor_Type myIntSens = createInstance(IntelligentSensor_Type.class, "IntelligentSensor");
			
			//this.addNodeAndReference(objectsFolder, myIntSens, Identifiers.Organizes);
			myObjectsFolder.addComponent(myIntSens);
			
			String uid = reader.getUIDIsens();

			myIntSens.setBrowseName(QualifiedName.parseQualifiedName(uid));
			int numSens = reader.getNumSens();
			myIntSens.setNumSensors(UnsignedInteger.getFromBits(numSens));
			int numAlgs = reader.getNumAlg();
			myIntSens.setNumAlgorithms(UnsignedInteger.getFromBits(numAlgs));
			
			
			int port = 9800;
			// Add solid sensors to intelligent sensor in the information model
			for(int i = 0; i<numSens; i++ ){				
				Sensor_Type sensorOPC = createSensorInstance(i, port);
				port++;
				myIntSens.addComponent(sensorOPC);
			}
			
			// Add available algorithms to intelligent sensor in the information model
			for(int i = 0; i<numAlgs; i++ ){
				Algorithm_Type algortihm = createAlgorithmInstance(i);			
				myIntSens.addComponent(algortihm);
			}			
			
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("a1");
		}
	}
	//#####################################################################################
	/**
	 * Creates an OPC UA node for an elementary sensor.
	 * @param num
	 * @return
	 * @throws Exception
	 */
	private Sensor_Type createSensorInstance(int num, int port) throws Exception{
		
		// sensor information for OPC UA and a local object for data exchange
		Sensor_Type sensorOPC = createInstance(Sensor_Type.class, "Sens" + num);
		
		// Browse name
		String uid  = reader.getSensorUID(num);
		sensorOPC.setBrowseName(QualifiedName.parseQualifiedName(uid));
		sensorOPC.setDisplayName(LocalizedText.english(uid));
		
		// Classifier -------------------------------------------------------------
		// Physical Quantity
		String physQuant = reader.getSensorClassifierTerm(num, "Physical Quantity");
		sensorOPC.getClassifierNode().setPhysicalQuantity(physQuant);
		
		// FOI
		String foiTitle = reader.getSensorFOITitle(num);
		String foiHref = reader.getSensorFOIHref(num);
		String[] foi = new String[2];
		foi[0]= foiTitle;
		foi[1]= foiHref;
		sensorOPC.getClassifierNode().setFOI(foi);
		
		// Location (Currently hard coded)!!!!!!!!!!!!
		sensorOPC.getClassifierNode().setLocation("Demo Rolle");
		
		// Output -------------------------------------------------------------
		int dimension = reader.getSensorOutputDimension(num);
		sensorOPC.getOutputNode().setDimension(UnsignedInteger.getFromBits(dimension));
		
		int dataLength = reader.getSensorOutputDataLength(num);
		//System.out.println(dataLength);
		sensorOPC.getOutputNode().setDataLength(UnsignedInteger.getFromBits(dataLength));
		
		String dataType = reader.getSensorOutputDataType(num);
		sensorOPC.getOutputNode().setDataType(dataType);
		
		sensorOPC.getOutputNode().setPort(UnsignedInteger.getFromBits(port));
		
		String interfaceType = reader.getSensorInterfaceType(num);
		int interfaceChannel = reader.getSensorInterfaceChannel(num);
		
		sensorOPC.getInterfaceNode().setType(interfaceType);
		sensorOPC.getInterfaceNode().setChannel(UnsignedInteger.getFromBits(interfaceChannel));
						
		return sensorOPC;
	}
	//#####################################################################################
	/**
	 * Creates an OPC UA node for an algorithm
	 * @param num
	 * @return
	 * @throws Exception
	 */
	private Algorithm_Type createAlgorithmInstance(int num) throws Exception{
		Algorithm_Type algorithm = createInstance(Algorithm_Type.class, "Algorithm" + num);
		
		// Browse name
		String uid  = reader.getAlgorithmUID(num);
		//System.out.println(uid);
		algorithm.setBrowseName(QualifiedName.parseQualifiedName(uid));
		algorithm.setDisplayName(LocalizedText.english(uid));
		
		//System.out.println(algorithm.getBrowseName());
				
		// Output -------------------------------------------------------------
		int dimension = reader.getAlgorithmOutputDimension(num);
		UnsignedInteger.getFromBits(dimension);
		algorithm.getOutputNode().setDimension(UnsignedInteger.getFromBits(dimension));
		
		//String physQuant = reader.getAlgorithmOutputPhysQuant(num);
		
		// Input  -------------------------------------------------------------
		dimension = reader.getAlgorithmInputDimension(num);
		UnsignedInteger.getFromBits(dimension);
		algorithm.getInputNode().setDimension(UnsignedInteger.getFromBits(dimension));
		
		String[] physQuants = (String[]) reader.getAlgorithmInputPhysQuant(num);
		algorithm.getInputNode().setPhysicalQuantities(physQuants);
		
		
		// Path to source code ------------------------------------------------
		String path = reader.getAlgorithmCodePath(num);
		algorithm.setCodePath(path);
		
		return algorithm;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.prosysopc.ua.server.NodeManagerUaNode#init()
	 */
	@Override
	protected void init() throws StatusException, UaNodeFactoryException {
		super.init();

		try {
			createAddressSpace();
		} catch (XMLReaderException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (UaInstantiationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
