/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package opcua.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.opcfoundation.ua.builtintypes.LocalizedText;
import org.opcfoundation.ua.builtintypes.UnsignedInteger;
import org.opcfoundation.ua.core.ApplicationDescription;
import org.opcfoundation.ua.core.UserTokenPolicy;
import org.opcfoundation.ua.transport.security.HttpsSecurityPolicy;
import org.opcfoundation.ua.transport.security.KeyPair;
import org.opcfoundation.ua.transport.security.SecurityMode;
import org.opcfoundation.ua.utils.EndpointUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.prosysopc.ua.ApplicationIdentity;
import com.prosysopc.ua.CertificateValidationListener;
import com.prosysopc.ua.PkiFileBasedCertificateValidator;
import com.prosysopc.ua.SecureIdentityException;
import com.prosysopc.ua.StatusException;
import com.prosysopc.ua.UaAddress;
import com.prosysopc.ua.UaApplication.Protocol;
import com.prosysopc.ua.samples.server.*;
import com.prosysopc.ua.server.NodeManagerListener;
import com.prosysopc.ua.server.UaServer;
import com.prosysopc.ua.server.UaServerException;
import com.prosysopc.ua.server.UserValidator;
import fileManager.Read;
import fusion.manager.Algorithm;
import fusion.manager.FeatureStore;
import fusion.manager.Sensor;
import gpioreader.Pi4JUSBSer;

/**
 * An OPC UA server that represents an intelligent sensor that is incorporated into a 
 * Sensor and Information Fusion System. The intelligent sensor locally contains a
 * self-description that is read and shared with an orchestration system. The latter
 * carries out automated feature orchestration for Sensor and Information Fusion Systems.
 */
public class ServerMain{

	protected static String APP_NAME = "IntelligentSensor";
	protected static String discoveryServerUrl;
	private static String deviceIP;
	private static boolean initSens = false;
	private static Read reader;
	private FeatureStore featureStore;
	
	// Port for data exchange.
	// !!!!!!! Has to be incremented after usage !!!!!!!
	int port = 9800;
	
	//required for serial usb interface. Can only be instantiated once
	boolean isInitUSB=false;
	Pi4JUSBSer usbSerial;
	
	private static String folderNameDescriptionFiles;
	private static boolean isSetFolderName = false;
	
	List<Sensor> sensors = new ArrayList<Sensor>();
	List<Algorithm> algorithms = new ArrayList<Algorithm>();
	private static long startTime;
	private static long endTime;
	private static long elapsed;
	// *******************************************************************************************************************
	/**
	 * main routine
	 * 
	 * @param args
	 *            command line arguments
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		startTime = System.currentTimeMillis();

		//Switch off all log4j logging
		PropertyConfigurator.configure("log.properties");
		Logger.getRootLogger().setLevel(Level.OFF);
		Logger.getLogger("org.opcfoundation.ua").setLevel(Level.OFF);
		Logger.getLogger("com.prosysopc.ua").setLevel(Level.OFF);

		
		try {
			if (!parseCmdLineArgs(args)) {
				usage();
				System.exit(0);		
			}
		} catch (IllegalArgumentException e) {
			System.out.println("Invalid cmd line argument: " + e.getMessage());
			usage();
			System.exit(0);		
		}
		
		//stop application if there is no folder defined that includes the self-description
		if(!isSetFolderName){
			System.out.println("!!!Please set folder name of self-description!!!\n");
			usage();
			System.exit(0);			
		}
		
		// creater serverMain instance
		ServerMain serverMain = new ServerMain();
		// get local IP
		deviceIP = getIP();
		
		//Inititialise reader for description files
		reader = new Read(folderNameDescriptionFiles);
		
		//Get uid of intelligent sensor
		String uid = reader.getUIDIsens();
		
		//init OPC UA Server
		serverMain.initialize(52520, 52443, APP_NAME, deviceIP, uid);
	
		// Initialize sensor objects. Also the thread for sensor data, iff initSens==true
		serverMain.initSensorInterfaces();
		
		//Initialize feature store and its config thread
		//serverMain.readAlgorithms();
		serverMain.featureStore = new FeatureStore(serverMain.algorithms);
		serverMain.featureStore.initConfigThread();
				
		// Create address space
		serverMain.createAddressSpace();

		// Start server
		serverMain.run(false);
		
		if(initSens){
			serverMain.closeSensorInterfaces();
			//serverMain.usbSerial.stop();
		}
		System.exit(0);		
	}
	
	// *******************************************************************************************************************
	/**
	 * Get IP of local device for domain 192.168.1.xxx
	 * @return
	 * @throws Exception
	 */
	private static String getIP() throws Exception{
		while(true){
			Enumeration<?> en = NetworkInterface.getNetworkInterfaces();
			while(en.hasMoreElements())
			{
			    NetworkInterface n = (NetworkInterface) en.nextElement();
			    Enumeration<?> ee = n.getInetAddresses();
			    while (ee.hasMoreElements())
			    {			    	
			        InetAddress i = (InetAddress) ee.nextElement();
			        if(!i.getHostAddress().contains(":") && //no IPv6
			        		!i.getHostAddress().substring(0, Math.min(i.getHostAddress().length(), 9)).equals("127.0.0.1")){ //no localhost
			        	return i.getHostAddress();
			        }		        	
			    }
			}
		}		
	}
	// *******************************************************************************************************************
	/**
	 * @param string
	 */
	public static void println(String string) {
		System.out.println(string);
	}
	// *******************************************************************************************************************
	/**
	 * @return
	 */
	private static String readInput() {
		BufferedReader stdin = new BufferedReader(new InputStreamReader(
				System.in));
		String s = null;
		do
			try {
				s = stdin.readLine();
			} catch (IOException e) {
				e.printStackTrace();
			}
		while ((s == null) || (s.length() == 0));
		return s;
	}
	// *******************************************************************************************************************
	/**
	 * Parse Command line arguments. Expected options:
	 * -d url     			Define the DiscoveryServerUrl to register the application to
	 * -s         			Initialize interface for sensor data access (GPIO, UART, USB (Serial))
	 * -desc 'folder name'	Set folder name that includes the self-description. The folder has to be located in 'System' directory, next to the executable jar file.
	 * -? ; -h ; -help		Show this help text
	 * 
	 *
	 * @param args
	 * @return
	 */
	protected static boolean parseCmdLineArgs(String[] args)
			throws IllegalArgumentException {
		int i = 0;
		while ((args.length > i)
				&& ((args[i].startsWith("-") || args[i].startsWith("/")))) {
			if (args[i].equals("-d"))
				discoveryServerUrl = "opc.tcp://" + args[++i] + ":4840";
			else if (args[i].equals("-s"))
				initSens = true;
			else if (args[i].equals("-desc")){
				folderNameDescriptionFiles = args[++i];
				isSetFolderName = true;
			}
			else if (args[i].equals("-?") || args[i].equals("-help") || args[i].equals("-h"))
				return false;
			else
				throw new IllegalArgumentException(args[i]);
			i++;
		}
		return true;
	}
	// *******************************************************************************************************************
	protected static void usage() {
		System.out.println("Usage: sudo java -jar APPNAME.jar [-s] [-d] [url] [-desc] [folder name]");
		System.out.println("   -d url     			Define the DiscoveryServerUrl to register the application to");
		System.out.println("   -s         			Initialize interface for sensor data access (GPIO, UART, USB (Serial))");
		System.out.println("   -desc 'folder name'	Set folder name that includes the self-description. The folder has to be located in 'System' directory, next to the executable jar file.");
		System.out.println("   -? ; -h ; -help		Show this help text");
		System.out.println("");
	}
	
	protected IntelligentSensorNodeManager ismn;
	protected NodeManagerListener myNodeManagerListener = new MyNodeManagerListener();
	protected UaServer server;
	protected final UserValidator userValidator = new MyUserValidator();
	protected final CertificateValidationListener validationListener = new MyCertificateValidationListener();

	public UaServer getServer() {
		return server;
	}

	// *******************************************************************************************************************
	/**
	 * Create a sample address space with a new folder, a device object, a level
	 * variable, and an alarm condition.
	 *
	 * @throws Exception 
	 *
	 */
	protected void createAddressSpace() throws Exception {
		// Load the standard information models
		loadInformationModels();

		ismn = new IntelligentSensorNodeManager(server, MyNodeManager.NAMESPACE, reader);
		ismn.addListener(myNodeManagerListener);
		ismn.getIoManager().addListeners(new MyIoManagerListener());
		

	}

	// *******************************************************************************************************************
	protected void initialize(int port, int httpsPort, String applicationName, String applicationIP, String uid)
			throws SecureIdentityException, IOException, UaServerException {

		// Create server
		server = new UaServer();

		// Use PKI files to keep track of the trusted and rejected client
		// start certificate validator
		final PkiFileBasedCertificateValidator validator = new PkiFileBasedCertificateValidator();
		server.setCertificateValidator(validator);
		validator.setValidationListener(validationListener);

		// *** Application Description is sent to the clients
		ApplicationDescription appDescription = new ApplicationDescription();
		appDescription.setApplicationName(new LocalizedText(applicationName,
				Locale.ENGLISH));
		appDescription.setApplicationUri("urn:"+applicationIP+":OPCUA:"
				+ applicationName);
		appDescription.setProductUri(uid);
		
		String[] urls = server.getServerUris();
		for(int j= 0; j<urls.length;j++){
			System.out.println("----\n"+urls[j]);
		}

		//set ports for opc server
		server.setPort(Protocol.OpcTcp, port);
		server.setPort(Protocol.Https, httpsPort);
		server.setServerName("OPCUA/" + applicationName);

		//only IPv4 or IPv6 allowed (depending on isEnableIPv6)
		server.setBindAddresses(EndpointUtil.getInetAddresses(server
				.isEnableIPv6()));

		// Certificates

		File privatePath = new File(validator.getBaseDir(), "private");
		KeyPair issuerCertificate = ApplicationIdentity.loadOrCreateIssuerCertificate("ProsysSampleCA", privatePath, "opcua", 3650, false);
		final ApplicationIdentity identity = ApplicationIdentity.loadOrCreateCertificate(appDescription, "Sample Organisation", "opcua", privatePath, null, null, true);
		String hostName = ApplicationIdentity.getActualHostName();
		identity.setHttpsCertificate(ApplicationIdentity.loadOrCreateHttpsCertificate(appDescription, hostName,"opcua", issuerCertificate, privatePath, true));
		server.setApplicationIdentity(identity);
		server.getHttpsSettings().setCertificateValidator(validator);

		// Security
		server.setSecurityModes(SecurityMode.NON_SECURE);
		server.getHttpsSettings().setHttpsSecurityPolicies(HttpsSecurityPolicy.ALL);

		// tokens
		server.addUserTokenPolicy(UserTokenPolicy.ANONYMOUS);
		server.addUserTokenPolicy(UserTokenPolicy.SECURE_USERNAME_PASSWORD);
		server.addUserTokenPolicy(UserTokenPolicy.SECURE_CERTIFICATE);
		server.setUserValidator(userValidator);
		
		// Register on the local discovery server
		try {
			UaAddress discoveryAddress = new UaAddress(discoveryServerUrl);
			server.setDiscoveryServerAddress(discoveryAddress);
		} catch (URISyntaxException e) {
			System.out.println("URL of discovery server is not valid " + e.getMessage());
		}
		
		//init server
		server.init();

		//mxaimal session count
		server.getSessionManager().setMaxSessionCount(500);
		//timeout = 2 hrs
		server.getSessionManager().setMaxSessionTimeout(7200000);
		server.getSubscriptionManager().setMaxSubscriptionCount(50);

	}
	// *******************************************************************************************************************
	/**
	 * Reads the description file of elementary sensors and, generates Sensor-Objects, and initializes the data interfaces. 
	 * Interfaces are currently implemented as Threads (see classes Sensor and SensorThread).
	 */
	protected void initSensorInterfaces(){		
		try {
			
			//String uid = reader.getUIDIsens();
			int numSens = reader.getNumSens();
			int numAlgs = reader.getNumAlg();
			
			
			// Add solid sensors to intelligent sensor in the information model
			for(int i = 0; i<numSens; i++ ){
				// Browse name
				String uidS  = reader.getSensorUID(i);
				
				// Classifier -------------------------------------------------------------
				// Physical Quantity
				String physQuant = reader.getSensorClassifierTerm(i, "Physical Quantity");
				
				// FOI
				String foiTitle = reader.getSensorFOITitle(i);
				String foiHref = reader.getSensorFOIHref(i);
				String[] foi = new String[2];
				foi[0]= foiTitle;
				foi[1]= foiHref;				
				
				// Output -------------------------------------------------------------
				int dimension = reader.getSensorOutputDimension(i);
				int dataLength = reader.getSensorOutputDataLength(i);
				UnsignedInteger.getFromBits(dimension);
				
				String interfaceType = reader.getSensorInterfaceType(i);
				int interfaceChannel = reader.getSensorInterfaceChannel(i);
				
				// Generate sensor object
				Sensor dummy = new Sensor(uidS, physQuant, dimension, "int", dataLength, getIP(),
						interfaceType, interfaceChannel, port, foiHref);
				
				
				if(initSens){
					if(interfaceType.equals("USB")){
						//Instantiate serial usb object only once and use it for all 
						// solid sensors that are attached to this interface
						
						if(!isInitUSB){
							usbSerial = new Pi4JUSBSer();
						}
						dummy.initUSBDataInterface(usbSerial);
						isInitUSB=true;
					}
					else{
						dummy.initDataInterface();
					}
					
				}
				sensors.add(dummy);
				port ++;
			}
			
			
			
			// Add solid sensors to intelligent sensor in the information model
			for(int i = 0; i<numAlgs; i++ ){
				// Browse name
				String uidS  = reader.getAlgorithmUID(i);
				
				// Classifier -------------------------------------------------------------
				// Physical Quantity
				String[] physQuant = (String[]) reader.getAlgorithmInputPhysQuant(i);
				
				// Input
				int inDimension = reader.getAlgorithmInputDimension(i);
				
				// Output -------------------------------------------------------------
				int outDimension = reader.getAlgorithmOutputDimension(i);
				//int outDataLength = reader.getSensorOutputDataLength(i);
				
				// Generate algorithm object
				Algorithm dummy = new Algorithm(uidS, physQuant, inDimension, outDimension, "double", 8);
				
				algorithms.add(dummy);
				
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		//System.out.println(sensors.size());
	}
	// *******************************************************************************************************************
	/**
	 * Reads the description file of algorithms, generates Algorithm-Objects, and initializes the data interfaces. 
	 * Interfaces are currently implemented as Threads (see classes Sensor and SensorThread).
	 */
	protected void readAlgorithms(){		
		try {
			
			//String uid = reader.getUIDIsens();
			int numAlgs = reader.getNumAlg();
			//int numAlgs = reader.getNumAlg();
			
			
			// Add solid sensors to intelligent sensor in the information model
			for(int i = 0; i<numAlgs; i++ ){
				// Browse name
				String uidS  = reader.getAlgorithmUID(i);
				
				// Classifier -------------------------------------------------------------
				// Physical Quantity
				String[] physQuant = (String[]) reader.getAlgorithmInputPhysQuant(i);
				
				// Input
				int inDimension = reader.getAlgorithmInputDimension(i);
				
				// Output -------------------------------------------------------------
				int outDimension = reader.getAlgorithmOutputDimension(i);
				//int outDataLength = reader.getSensorOutputDataLength(i);
				
				// Generate algorithm object
				Algorithm dummy = new Algorithm(uidS, physQuant, inDimension, outDimension, "double", 8);
				
				algorithms.add(dummy);
				
			}			
		} catch (Exception e) {
			e.printStackTrace();
			
		}
	}
	// *******************************************************************************************************************
	private void closeSensorInterfaces(){		
		for(int i=0; i < sensors.size(); i++){
			sensors.get(i).closeSensorInterface();
		}
	}
	// *******************************************************************************************************************
	/**
	 * Write semantics to an XML-File for Profinet configurations
	 */
	public void generateProfinetXML(){
		try {			
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
			// root elements
			Document doc = docBuilder.newDocument();
			Element root = doc.createElement("DevParam");
			doc.appendChild(root);
			
			Element addr = doc.createElement("IP");
			String dummy = getIP();
			String[] addrS = dummy.split("\\.");
			addr.setAttribute("d1", addrS[0]);
			addr.setAttribute("d2", addrS[1]);
			addr.setAttribute("d3", addrS[2]);
			addr.setAttribute("d4", addrS[3]);
			root.appendChild(addr);
			

			Element mask = doc.createElement("Mask");
			mask.setAttribute("d1", "255");
			mask.setAttribute("d2", "255");
			mask.setAttribute("d3", "255");
			mask.setAttribute("d4", "0");
			root.appendChild(mask);
			
			Element modules = doc.createElement("Modules");
			root.appendChild(modules);
			for(int i = 0; i<sensors.size();i++){
				Element module = doc.createElement("Module");
				module.setAttribute("type", "input");
				module.setAttribute("datalength", String.valueOf(sensors.get(i).getDataLength()));
				module.setAttribute("slot", String.valueOf(i+1));
				modules.appendChild(module);
			}
						
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			//transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File(reader.getNameIsens() + ".xml"));
			
			transformer.transform(source, result);
			
		}catch(Exception e){
			System.out.println("Exception in XML-writing: " + e.getMessage());
		}
	}

	/**
	 * Load information models 
	 * how to create information models; see codegen/Readme.md 
	 */
	protected void loadInformationModels() {	
		server.registerModel(opcua.types.server.InformationModel.MODEL);
		try {
			server.getAddressSpace().loadModel(new File(System.getProperty("user.dir"),"codegen/intelligentsensor.xml").toURI());
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		
	}

	// *******************************************************************************************************************
	/**
	 * Run the server.
	 *
	 * @param enableSessionDiagnostics
	 * @throws UaServerException
	 * @throws StatusException
	 */
	protected void run(boolean enableSessionDiagnostics)
			throws UaServerException, StatusException {
		server.start();	
		
		/******************************************************************************/
		// Wait for user input
		boolean wait = true;
		do {
			endTime = System.currentTimeMillis();
	        
	        elapsed = (endTime - startTime)/1000;
	        System.out.println("Elapsed Time: " + elapsed + "s");
			try {
				 System.out.println("Press 'c' to shutdown the server.");
				String inp = readInput();
				if(inp.equals("c")){
					wait = false;
				}
				
			} catch (Exception e) {
				e.printStackTrace();
			}

		} while (wait);
		/******************************************************************************/
		// Notify the clients about a shutdown, with a 3 second delay
		System.out.println("Shutting down...");
		server.shutdown(3, new LocalizedText("Closed by user", Locale.ENGLISH));
		this.featureStore.closeConfigThread();
	}

}
