/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.Semaphore;

import fusion.manager.aggregation.BalTLCS;
import fusion.manager.aggregation.IIWOWA;

/**
 * A single Attribute that consist of a UID, physical phenomena, and a list of
 * suitable features.
 * 
 * @author A_Fritze
 *
 */
public abstract class Attribute {

	String UID;
	List<String> suitablePhysPheno;
	List<String> suitableFeatures;
	List<Feature> features;
	List<String> sensors;

	// socket server port on which it will listen
	private List<Integer> portsIn;
	private List<AttributeInThread> inThreads;

	private BalTLCS baltlcs = new BalTLCS();
	private List<Double> BBAs = new ArrayList<Double>();
	//ConcurrentLinkedQueue<List<Double>> attributeHistory = new ConcurrentLinkedQueue<List<Double>>();
	//private LinkedList<List<Double>> attributeHistory = (LinkedList<List<Double>>) Collections.synchronizedList(new LinkedList<List<Double>>());
	CircularArrayList<List<Double>> attributeHistory = new CircularArrayList<>(20);
	private boolean initAttributeHistory=false;
	private IIWOWA iiwowa;
	private Semaphore iiwowaSemaphore;
	private Semaphore historySemaphore = new Semaphore(1, true);
	
	// *******************************************************************************************************************
	/**
	 * The index is required to define the attributes input array position of
	 * the IIWOWA operator
	 */
	private int index;

	// Counter for output at console
	private int countOut = 0;

	// *******************************************************************************************************************
	/**
	 * 
	 * @param UID
	 * @param suitablePhysPheno
	 * @param suitableFeatures
	 */
	public Attribute(String UID, List<String> suitablePhysPheno, List<String> suitableFeatures, Semaphore iiwowaSemaphore) {

		this.UID = UID;
		this.suitablePhysPheno = suitablePhysPheno;
		this.suitableFeatures = suitableFeatures;
		this.iiwowaSemaphore = iiwowaSemaphore;

		sensors = new ArrayList<String>();
		features = new ArrayList<Feature>();

		portsIn = new ArrayList<Integer>();
		inThreads = new ArrayList<AttributeInThread>();

	}
	// *******************************************************************************************************************
	/**
	 * Sets a new BBA at position "id" of the BBA-List. This method is called by
	 * the "AttributInThread" if a new BBA is available for one feature. Iff
	 * values for each input of this attribute are available the BalTLCS is
	 * executed to compute the attribute health. The IIWOWA operator is also
	 * called but only executed iff attribute health's of every attribute are
	 * available.
	 * 
	 * @param ind
	 * @param value
	 */
	public void setBBAandFuse(int ind, Double value) {

		// Set BBA value at index (ind)
		BBAs.set(ind, value);
		//System.out.println("BBAs: " + BBAs);
		//System.out.println("BBAs:  " + BBAs);
		if (!BBAs.contains(null)) {
			// System.out.println("###");
			// Compute attribute health
			countOut++;
			// System.out.println("BBAs " + BBAs);
			List<Double> result = baltlcs.fuse(BBAs);
			
			//Only write every tenth result
			//!!!Also IIWOWA is applied every tenth time!!!
			try {
				historySemaphore.acquire();
				//Print out results
				if (countOut >= 10) {
					try {
						iiwowaSemaphore.acquire();
						iiwowa.setValuesAndFuse(result.get(0), result.get(2), index);
						iiwowaSemaphore.release();
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					countOut = 0;
				}
				attributeHistory.addCirc(result);
				historySemaphore.release();
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			
			
			/*
			//Print
			System.out.println(UID + ": ");
			for(List<Double> values : attributeHistory){
				System.out.print(values.get(0) + ";");
			}
			*/		
			
			//System.out.println("Attribute " + UID + " :" + result.get(0));
			//System.out.println("-------");
			
			/*
			// Set mass and importance of the attribute for the IIWOWA operator
			// at position "index"
			try {
				iiwowaSemaphore.acquire();
				iiwowa.setValuesAndFuse(result.get(0), result.get(2), index);
				iiwowaSemaphore.release();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			*/
			
		}
	}

	// ################################################################################################
	// Getter
	public String getUID() {
		return UID;
	}
	// *******************************************************************************************************************
	/**
	 * Returns a String List of suitable physical phenomena for this attribute
	 * 
	 * @return
	 */
	public List<String> getSuitablePhysPheno() {
		return suitablePhysPheno;
	}
	// *******************************************************************************************************************
	/**
	 * Returns a String List of suitable features for this attribute (e.g. mean,
	 * variance, etc)
	 * 
	 * @return
	 */
	public List<String> getSuitableFeatures() {
		return suitableFeatures;
	}
	// *******************************************************************************************************************
	/**
	 * Returns the list of features that are associated to this attribute
	 * 
	 * @return
	 */
	public List<Feature> getFeatures() {
		return features;
	}
	// *******************************************************************************************************************
	/**
	 * Returns a String List of the UIDs of all features
	 * 
	 * @return
	 */
	public List<String> getFeaturesStringList() {
		List<String> featuresString = new ArrayList<String>();

		for (int i = 0; i < features.size(); i++)
			featuresString.add(features.get(i).getUID());
		return featuresString;
	}
	// *******************************************************************************************************************
	/**
	 * Returns the list of sensors that are associated to the attribute
	 * 
	 * @return
	 */
	public List<String> getSensors() {
		return sensors;
	}
	// *******************************************************************************************************************
	/**
	 * Add feature to attribute
	 * 
	 * @param feature
	 */
	public void addFeature(Feature feature) {
		features.add(feature);
	}
	// *******************************************************************************************************************
	/**
	 * Add sensor to attribute
	 * 
	 * @param UID
	 */
	public void addSensor(String UID) {
		sensors.add(UID);
	}
	// *******************************************************************************************************************
	/**
	 * Add an input port to the attribute
	 * 
	 * @param port
	 */
	public void addInputPort(int port) {
		portsIn.add(port);
	}
	// *******************************************************************************************************************
	/**
	 * Return List of input ports of this attribute
	 */
	public List<Integer> getInputPorts() {
		return portsIn;
	}
	// *******************************************************************************************************************
	/**
	 * Set the IIWOWA operator for this application
	 * 
	 * @param iiwowa
	 */
	public void setIIWOWA(IIWOWA iiwowa) {
		this.iiwowa = iiwowa;
	}
	// *******************************************************************************************************************
	/**
	 * Set the index of this Atrribute for fusion using the IIWOWA operator
	 */
	public void setIndex(int ind) {
		index = ind;
	}
	// *******************************************************************************************************************
	/**
	 * Initialize Inputs the attribute.
	 */
	public void initThreads() {
		System.out.println(UID);
		Semaphore semaphor = new Semaphore(1, true);

		for (int i = 0; i < portsIn.size(); i++) {
			System.out.println("Open Atrribute Input Port: " + portsIn.get(i));
			inThreads.add(new AttributeInThread(portsIn.get(i), this, i, semaphor));
			inThreads.get(i).start();

			// initialise list of BBAs.
			BBAs.add(null);
		}
	}
	// *******************************************************************************************************************
	/**
	 * Close input threads of an attribute
	 */
	public void closeThreads() {
		for (int i = 0; i < inThreads.size(); i++) {
			inThreads.get(i).closeThread();
		}
	}
	// *******************************************************************************************************************
	/**
	 * Returns the last 20 results of the attribute health
	 * @return
	 */
	public CircularArrayList<List<Double>> getHistory(){
		CircularArrayList<List<Double>> attributeHistoryReturn = new CircularArrayList<>(20);
		
		try {
			historySemaphore.acquire();
			for(int i = 0; i< attributeHistory.size(); i++){
				attributeHistoryReturn.addCirc(attributeHistory.get(i));
			}
			historySemaphore.release();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return attributeHistoryReturn;
	}
}
