/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;

import fusion.manager.aggregation.IIWOWA;

/**
 * This class manages attributes for the Sensor and Information Fusion System.
 * 
 * @author A_Fritze
 *
 */
public class AttributeStore {
		
	//Lists of attributes
	private List<AttributeModule> AttributesM;
	private List<AttributeGeneral> AttributesG;
	
	//IIWOWAO operator
	IIWOWA iiwowa;
	
	private writeForUIThread writeForUI;
	
		
	public AttributeStore(){
		AttributesM = new ArrayList<AttributeModule>();
		AttributesG = new ArrayList<AttributeGeneral>();
		
		//writeForUI = new writeForUIThread(this);
	}	
	
	//###############################################################################################
	public List<AttributeModule> getAttributeModuleList(){
		return AttributesM;
	}	
	public AttributeModule getAttributeModule(String module){
		for(int i=0; i<AttributesM.size(); i++){
			if(AttributesM.get(i).getFoi().equals(module)){
				return AttributesM.get(i);
			}
		}		
		return null;
	}
	//###############################################################################################
	public List<AttributeGeneral> getAttributeGeneralList(){
		return AttributesG;
	}
	//###############################################################################################
	public void setAttributeModuleList(List<AttributeModule> newAttributeModuleList){
		AttributesM = newAttributeModuleList;
	}
	//###############################################################################################
	public void setAttributeGeneralList(List<AttributeGeneral> newAttributeGeneralList){
		AttributesG = newAttributeGeneralList;
	}	
	//###############################################################################################
	public void addAttributeModule(AttributeModule attribute){
		AttributesM.add(attribute);
	}
	//###############################################################################################
	public void addAttributeGeneral(AttributeGeneral attribute){
		AttributesG.add(attribute);
	}
	//###############################################################################################
	/**
	 * Initialise the IIWOWA operator according to the number of available attributes (n).
	 * Weights are set to 1/n
	 */
	public void initIIWOWA(){		
		List<Double> weights = new ArrayList<Double>();
		
		int numAttributes = AttributesM.size() + AttributesG.size();
		
		for(int i =0; i<numAttributes; i++){
			weights.add((1/(double)numAttributes));
		}		
		iiwowa = new IIWOWA(weights);
		
		int index = 0;
		for(int i=0; i<AttributesG.size(); i++){
			AttributesG.get(i).setIndex(index);
			AttributesG.get(i).setIIWOWA(iiwowa);
			index++;
		}
		for(int i=0; i<AttributesM.size(); i++){
			AttributesM.get(i).setIndex(index);
			AttributesM.get(i).setIIWOWA(iiwowa);
			index++;
		}
		
	}
	
	//###############################################################################################
	/** 
	 * Remove Attributes with single features.
	 */
	public void removeBadAttributes(){
		
		List<AttributeModule> module = new ArrayList<AttributeModule>();
		List<AttributeGeneral> general = new ArrayList<AttributeGeneral>();
		
		for(int i = 0; i < AttributesG.size(); i++){
			if(AttributesG.get(i).getFeatures().size() > 1){
				general.add(AttributesG.get(i));
			}
		}
		
		for(int i = 0; i < AttributesM.size(); i++){
			if(AttributesM.get(i).getFeatures().size() > 1){
				module.add(AttributesM.get(i));
			}
		}
		
		AttributesG = general;
		AttributesM = module;
	}
	//###############################################################################################
	/** 
	 * Close attribute threads
	 */
	public void closeAttributeThreads(){
		
		for(int i = 0; i < AttributesG.size(); i++){
			AttributesG.get(i).closeThreads();
		}
		
		for(int i = 0; i < AttributesM.size(); i++){
			AttributesM.get(i).closeThreads();
		}
	}
	//###############################################################################################
	/**
	 * Start write thread for csv file for UI
	 */
	public void startWriteCsvThread(){
	//	writeForUI.start();
	}
	//###############################################################################################
	/**
	 * Stop write thread for csv file for UI
	 */
	public void stoptWriteCsvThread(){
	//	if(writeForUI.isAlive()){
	//		writeForUI.shutdown();
	//		writeForUI.interrupt();
	//	}	
	}	
}

class writeForUIThread extends Thread{
	private boolean stop= false;
	AttributeStore store;    
    
	public writeForUIThread(AttributeStore store)
	{
		this.store=store;
	}
	
	public void run(){
		while(!stop){
			writeToCsv();
	    	 try {
	             writeForUIThread.sleep(2000);
	         } catch (Exception e) {
	             //System.out.println(e);
	         }
		}
		this.interrupt();
    }
	
	public void shutdown(){
		stop = true;
	}
	
	//###############################################################################################
	/**
	 * Writes the last (20) results of attribute health and system health
	 */
	public void writeToCsv(){
        try {
        	//System.out.println("Write");
        	File file = new File("/home/pi/Profinet/fusionData.csv");
    		file.createNewFile();
    		
			//System.out.println("Created file: " + file.getAbsolutePath());
	        FileWriter f1 = new FileWriter(file.getAbsoluteFile());
	        BufferedWriter bufferWriter = new BufferedWriter(f1);
	        
	        //Write Attribute data***************************************
	        for(AttributeModule attr : store.getAttributeModuleList()){
				CircularArrayList<List<Double>> history = attr.getHistory();
				if(!history.isEmpty()){
					String[] parsed = attr.UID.split(":");
					String name = "";
					for(int i=0; i<parsed.length;i++){
						if(i>=2)
							name = name + " " + parsed[i];
					}
					bufferWriter.write("attribute;" + name + ";");
					//System.out.println("## attribute;" + name + ";");
					for(List<Double> values : history){
						bufferWriter.write(values.get(0) + ";");
						//System.out.print(values.get(0) + ";");
					}
					//System.out.print("##\n");
					bufferWriter.write("\n");
				}
			}
	        for(AttributeGeneral attr : store.getAttributeGeneralList()){
				CircularArrayList<List<Double>> history = attr.getHistory();
				if(!history.isEmpty()){
					String[] parsed = attr.UID.split(":");
					String name = "";
					for(int i=0; i<parsed.length;i++){
						if(i>=2)
							name = name + " " + parsed[i];
					}
					bufferWriter.write("attribute;" + name + ";");
					//System.out.println("## attribute;" + name + ";");
					for(List<Double> values : history){
						bufferWriter.write(values.get(0) + ";");
						//System.out.print(values.get(0) + ";");
					}
					//System.out.print("##\n");
					//System.out.print("\n");
					bufferWriter.write("\n");
				}
			}
	        //Write System Health***************************************
	        CircularArrayList<Double> healths = store.iiwowa.getHistory();
	        if(!healths.isEmpty()){
	        	bufferWriter.write("system health;");
	        	bufferWriter.write("system health;");
	        	//System.out.println("## System Health: ");
	        	for(Double value : healths){
	        		bufferWriter.write(value + ";");
	        		//System.out.print(value + ";");
	        	}
	        	//System.out.print("##\n");
	        	bufferWriter.write("\n");
	        }
	        bufferWriter.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}



