/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager;

import java.util.ArrayList;
import java.util.List;

public class Feature implements FeatureClientInterface {

	String UID;
	List<String> suitablePhysPheno;
	List<Sensor> sensors;
	Algorithm algorithm;
	String foi;
	String ip;
	int portsIn;
	int portsOut;
	List<Integer> attributeInPorts;
	
	public Feature(String UID, List<String> suitablePhysPheno, Algorithm algorithm, String observedEntity){

		//Semantics
		this.UID = UID;
		this.suitablePhysPheno = suitablePhysPheno;
		this.algorithm = algorithm;
		this.foi = observedEntity;
		this.ip = algorithm.getIP();
		attributeInPorts = new ArrayList<Integer>();
		
		//Fusion
		sensors = new ArrayList<Sensor>();
	}
	
	//################################################################################################
	/**
	 * Get the UID of this feature
	 */
	public String getUID(){
		return UID;
	}
	
	/**
	 * Returns the list of physical phenomena that are suited for this feature
	 * @return
	 */
	public List<String> getSuitablePhysPheno(){
		return suitablePhysPheno;
	}
	
	/**
	 * Returns the list of solid sensors that are incorporated for this feature
	 */
	public List<Sensor> getSensors(){
		return sensors;
	}
	
	/**
	 * Returns feature of interest (observed entity)
	 */
	@Override
	public String getfoi() {
		return foi;
	}
	
	/**
	 * Returns the algorithm object that computes the value of this feature
	 */
	@Override
	public Algorithm getAlgorithm() {
		return algorithm;
	}
	
	/**
	 * Set the observed entity (feature of interest) for this feature
	 */
	@Override
	public void setfoi(String observedEntity) {
		foi = observedEntity;
	}
	
	/**
	 * Add sensor objects to feature
	 */
	public void addSensor(Sensor sensor){
		sensors.add(sensor);
	}

	/**
	 * Set I/O ports for socket communication of this feature
	 */
	@Override
	public void setPorts(int portsIn, int portsOut) {
		this.portsIn = portsIn;
		this.portsOut = portsOut;
	}

	/**
	 * Returns the input ports of this feature
	 */
	@Override
	public int getInPort() {
		return portsIn;
	}

	/**
	 * Returns the output ports of this feature
	 */
	@Override
	public int getOutPort() {
		return portsOut;
	}

	/**
	 * Returns the IP of the intelligent sensor the feature is located
	 */
	@Override
	public String getIP() {
		return ip;
	}

	/**
	 * An attribute has n inputs (membership values of features). For each input a port is required.
	 * The method returns the list of attribute input ports of this feature.
	 */
	@Override
	public List<Integer> getAttributePorts() {
		return attributeInPorts;
	}

	/**
	 * An attribute has n inputs (membership values of features). For each input a port is required.
	 * The method adds an input port to the list of all ports of an attribute.
	 */
	@Override
	public void addAttributePort(int attributePort) {
		attributeInPorts.add(attributePort);		
	}
}
