/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.manager.aggregation;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;

import fusion.manager.CircularArrayList;

public class IIWOWA {

	// *******************************************************************************************************************
	// Required variables for computation of the IIWOWA-Result
	private List<Double> weights = new ArrayList<Double>();
	private List<Double> masses = new ArrayList<Double>();
	private List<Double> importance = new ArrayList<Double>();
	private double result, orness, andness, iwowa, iwowa0, iwowa1;
	int n;
	int outCount = 0;
	//private ConcurrentLinkedQueue<Double> iiwowaHistory = new ConcurrentLinkedQueue<Double>();
	CircularArrayList<Double> iiwowaHistory = new CircularArrayList<>(20);
	
	private Semaphore historySemaphore = new Semaphore(1, true);

	// *******************************************************************************************************************
	/**
	 * Constructor of the Implicative Importance Weighted Ordered Weighted
	 * Averaging Operator.
	 * 
	 * @param weights
	 *            The operator is initialized by the weighting vector
	 */
	public IIWOWA(List<Double> weights) {
		this.weights = weights;
		for (int i = 0; i < weights.size(); i++) {
			masses.add(null);
			importance.add(null);
		}
		n = weights.size();
	}

	// *******************************************************************************************************************
	/**
	 * Implicative Importance Weighted Ordered Weighted Averaging Operator.
	 * 
	 * @param masses
	 * @param importance
	 * @return Averaging value in range [0...1]
	 */
	public Object setValuesAndFuse(double mass, double I, int index) {

		masses.set(index, mass);
		importance.set(index, I);
		// System.out.println("masses: "+ masses + " I: " + importance);

		if (!masses.contains(null)) {
			
			outCount++;
			orness = 0;
			andness = 0;
			computeOrness();
			andness = 1 - orness;
			iwowa0 = IWOWA0();
			iwowa1 = IWOWA1();
			iwowa = IWOWA();

			// Log.d("IIWOWA", "Masses: " + masses.toString());
			// Log.d("IIWOWA", "Imp: " + importance.toString());
			// Log.d("IIWOWA", "ANDNESS: " + String.valueOf(andness) + "\nI0: "
			// + String.valueOf(iwowa0) + "\nI1: " + String.valueOf(iwowa1) +
			// "\nI : " + String.valueOf(iwowa));

			// System.out.println("andness: "+ andness + " iwowa0: " + iwowa0 +
			// " iwowa1: " + iwowa1 + " iwowa: " + iwowa);
			// Norm
			result = (iwowa - iwowa0) / (iwowa1 - iwowa0);

			/*
			if (outCount >= 1000) {
				System.out.println("System Health = " + result + " " + masses
				 + " " + importance);
				outCount = 0;
			}
			*/

			try {
				historySemaphore.acquire();
				iiwowaHistory.addCirc(result);
				//System.out.println("System Health = " + result);
				historySemaphore.release();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
			//semaphor.release();
			//System.out.println("Fusion result: " + result);
			return result;
		}
		return null;
	}

	// *******************************************************************************************************************
	/**
	 * Importance Weighted Ordered Weighted Averaging Operator for all masses
	 * m=0
	 * 
	 * @return Averaging value
	 */
	private double IWOWA0() {
		double res = 0;
		double dummy = 0;
		List<Double> dummyL = new ArrayList<Double>();
		for (int i = 0; i < n; i++) {
			dummy = (andness * (1 - importance.get(i)));
			dummyL.add(dummy);
		}
		List<Double> dummyLO = order(dummyL);

		for (int i = 0; i < n; i++) {
			res += weights.get(i) * dummyLO.get(i);
		}

		return res;
	}

	// *******************************************************************************************************************
	/**
	 * Importance Weighted Ordered Weighted Averaging Operator for all masses
	 * m=1
	 * 
	 * @return Averaging value
	 */
	private double IWOWA1() {
		double res = 0;
		double dummy = 0;
		List<Double> dummyL = new ArrayList<Double>();
		for (int i = 0; i < n; i++) {
			dummy = (andness + (importance.get(i) * orness));
			dummyL.add(dummy);
		}
		List<Double> dummyLO = order(dummyL);

		for (int i = 0; i < n; i++) {
			res += weights.get(i) * dummyLO.get(i);
		}
		return res;
	}

	// *******************************************************************************************************************
	/**
	 * Importance Weighted Ordered Weighted Averaging Operator
	 * 
	 * @return
	 */
	private double IWOWA() {
		double res = 0;
		double dummy = 0;
		List<Double> dummyL = new ArrayList<Double>();
		for (int i = 0; i < n; i++) {
			dummy = (andness + (importance.get(i) * (masses.get(i) - andness)));
			dummyL.add(dummy);
		}
		List<Double> dummyLO = order(dummyL);

		for (int i = 0; i < n; i++) {
			res += weights.get(i) * dummyLO.get(i);
		}
		return res;
	}

	// *******************************************************************************************************************
	/**
	 * Compute the orness of the operator
	 */
	private void computeOrness() {
		double sum = 0;
		for (int i = 1; i <= n; i++) {
			sum += ((double) n - (double) i) * weights.get(i - 1);
		}
		// System.out.println("sum: "+ sum + " n: " + n);
		orness = sum / (double) (n - 1);
	}

	// *******************************************************************************************************************
	/**
	 * Sort a List in decreasing order depending on the value of the list
	 * elements.
	 * 
	 * @param list
	 * @return List sorted in decreasing order
	 */
	private List<Double> order(List<Double> list) {
		for (int k = 0; k < list.size(); k++) {
			for (int l = k + 1; l < list.size(); l++) {

				if (list.get(k) < list.get(l)) {
					// Sort masses
					double dummy1 = list.get(k);
					list.set(k, list.get(l));
					list.set(l, dummy1);
				}
			}
		}
		return list;
	}

	// *******************************************************************************************************************
	/**
	 * Returns the last 20 results for the system health
	 * 
	 * @return
	 */
	public CircularArrayList<Double> getHistory() {
		CircularArrayList<Double> iiwowaHistoryReturn = new CircularArrayList<>(20);
		
		try {
			historySemaphore.acquire();
			for(int i =0; i<iiwowaHistory.size();i++){
				iiwowaHistoryReturn.addCirc(iiwowaHistory.get(i));
			}
			historySemaphore.release();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}	
		
		return iiwowaHistoryReturn;
	}
}
