/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package fusion.orchestration;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.Semaphore;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.vast.sensorML.SMLHelper;
import org.vast.sensorML.SMLUtils;
import org.vast.xml.XMLReaderException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import fusion.manager.Algorithm;
import fusion.manager.AttributeGeneral;
import fusion.manager.AttributeModule;
import fusion.manager.AttributeStore;
import fusion.manager.Feature;
import fusion.manager.FeatureStore;
import fusion.manager.IntelligentSensor;
import fusion.manager.Sensor;
import net.opengis.sensorml.v20.ClassifierList;
import net.opengis.sensorml.v20.PhysicalComponent;
import net.opengis.sensorml.v20.PhysicalSystem;
import net.opengis.sensorml.v20.Term;


/**
 * The Knowledge Base (KB) implements functionalities that are required for automated feature orchestration.
 * 
 * @author A_Fritze
 *
 */
public class KnowledgeBase{
	
	//SML Factory for management of self-description
	private SMLHelper smlHelper = new SMLHelper();
	private SMLUtils smlUtils = new SMLUtils(SMLUtils.V2_0);
	//private SWEHelper sweHelper = new SWEHelper();
	//private GMLFactory gmlHelper = new GMLFactory();
	
	//List of intelligent sensors
	private List<IntelligentSensor> listOfIntSens;
	//List of solid sensors
	private List<Sensor> listOfSolidSens;
	//List of algorithms
	private List<Algorithm> listOfAlgorithms;
	//List of modules of the observed system
	private List<PhysicalSystem> listOfModules;
	
	//List of features (combination of sensors and algorithms) !!! Merkmale keine features of interest (fois, module)
	private FeatureStore featureStore;
	//private List<Feature> listOfFeatures;
	
	//The Attribute store manages available attributes
	private AttributeStore attributeStore;
	
	//System Manager IP
	private String SysManagerIP;
	
	//Folder name of use case scenario.
	//Includes the csv file for functional attributes.
	private String folderNameUseCase;
	
	//################################# Constructor ###########################################
	public KnowledgeBase(String ip, String folderNameUseCase){
		//IP of the local device that implements the KB
		this.SysManagerIP = ip;
		
		this.folderNameUseCase = folderNameUseCase;
		
		//List of intelligent sensors
		listOfIntSens = new ArrayList<IntelligentSensor>();
		//List of solid sensors
		listOfSolidSens = new ArrayList<Sensor>();
		//List of features (combination of sensors and algorithms)
		//listOfFeatures = new ArrayList<Feature>();
		//List of algorithms
		listOfAlgorithms = new ArrayList<Algorithm>();
		//List of modules of the observed system
		listOfModules = new ArrayList<PhysicalSystem>();
		
		//Feature Store for management of all features
		featureStore = new FeatureStore();
		
		//Attribute Store for management of all attributes
		attributeStore = new AttributeStore();
	}
	
	//##########################################################################################
	//##########################################################################################	
	//##########################################################################################
	/**
	 * Assign features to attributes
	 */
	public void orchestrate(){
		
		orchestrateModuleAttributes();
		
		orchestrateGeneralAttributes();
		
		//attributeStore.removeBadAttributes();
		
		removeInsufficientAttributes();
		
		startAttributeThreads();
		
		printResults();
		
		writeResults();	
	}
	
	//##########################################################################################
	/**
	 * Reset lists of features and attributes
	 */
	public void clearLists(){
				
		//List of features (combination of sensors and algorithms)
		featureStore.clear(); //= new ArrayList<PhysicalComponent>();
		
		//Attribute Store for management of all attributes
		attributeStore.closeAttributeThreads();
		attributeStore.stoptWriteCsvThread();
		attributeStore = new AttributeStore();
		
		//List of modules of the observed system
		listOfModules = new ArrayList<PhysicalSystem>();
	}
	
	//##########################################################################################
	/** 
	 * Initialize Attributes. Depending on available modules and the physical phenomena of sensors.
	 * Functional attributes are manually added. 
	 */
	public void initAttributes(){
		
		addPhysQuantToModules();
		
		//Semaphore
		Semaphore iiwowaSemaphore = new Semaphore(1, true);
		
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Module Attributes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		//First generate Attributes for Modules
		for(int i=0; i < listOfModules.size(); i++){
			
			String urn = listOfModules.get(i).getIdentifier().getValue();
			String[] parsedUrn = urn.split(":");
			//String id = listOfModules.get(i).getId();
			
			String id = parsedUrn[parsedUrn.length - 1];
			//System.out.println("id: " + id);
			//System.out.println("urn: " + urn);
			
			//Add module attributes to attribute store. Modules have no particular phys phenomena or specific features
			if(!urn.equals("inIT:product")){
				attributeStore.addAttributeModule(new AttributeModule("inIT:attributes:" + id.toLowerCase(), urn, null, null, iiwowaSemaphore));
						
			//System.out.println("inIT:attributes:"+ id.toLowerCase() + " - " + title + " - " + urn + " - " + parentModule)
			
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Physical Quantities of Modules %%%%%%%%%%%%%%%%%%%%%%%%
				if(!listOfModules.get(i).getClassificationList().isEmpty()){ //Es muss mindestens ein solid sensor dem modul zugewiesen sein!!!!!
				for(int j = 0; j < listOfModules.get(i).getClassificationList().get(0).getClassifierList().size(); j++){
					String physPhen = listOfModules.get(i).getClassificationList().get(0).getClassifierList().get(j).getValue();
					
					attributeStore.addAttributeGeneral(new AttributeGeneral("inIT:attributes:" + id.toLowerCase() + ":" + physPhen ,"physical", urn,
							Arrays.asList(physPhen), null, iiwowaSemaphore));
				}
				}
			}
		}
		
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Quality Attributes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		//Add Attribute for printing quality
		attributeStore.addAttributeGeneral(new AttributeGeneral("inIT:attributes:quality","quality", "inIT:product",
				Arrays.asList("image"), Arrays.asList("image","BleedingsDetectionProcess", "HoleDetectionProcess", "color", "homogenity") ,iiwowaSemaphore));
		
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Functional Attributes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Stored in an external csv file %%%%%%%%%%%%%%%%%%%%%%%%
		String csvFile = "System/"+ folderNameUseCase + "/FunctionalAttributes.csv";
        BufferedReader br = null;
        String line = "";
        String cvsSplit1 = ";";
        String cvsSplit2 = ",";

        try {

            br = new BufferedReader(new FileReader(csvFile));
            while ((line = br.readLine()) != null) {

                // use comma as separator
                String[] attributeInfo = line.split(cvsSplit1);
                System.out.println(attributeInfo[0] + "---" + attributeInfo[1] + "---" + attributeInfo[2] + "---" + attributeInfo[3] + "---" + attributeInfo[4]);
                ArrayList<String> allowedPhenomena = new ArrayList<String>();
                ArrayList<String> allowdAlgorithms = new ArrayList<String>();
                
                String[] phenomenas = attributeInfo[3].split(cvsSplit2);
                String[] algos = attributeInfo[4].split(cvsSplit2);
                
                for(String value : phenomenas){
                	allowedPhenomena.add(value);
                }
                for(String value : algos){
                	allowdAlgorithms.add(value);
                }
                //System.out.println(attributeInfo[0] + "---" + attributeInfo[1] + "---" + attributeInfo[3] + "---" + allowedPhenomena + "---" + allowdAlgorithms);
                //generate functional attribute
                attributeStore.addAttributeGeneral(new AttributeGeneral(attributeInfo[0],attributeInfo[1],attributeInfo[2],allowedPhenomena,allowdAlgorithms,iiwowaSemaphore));
            }

        } catch (IOException e) {
            e.printStackTrace();
            System.out.println("#######Could not read functional attributes######");
        }
		/*
		attributeStore.addAttributeGeneral(new AttributeGeneral("inIT:attributes:running smoothness","functional", "inIT:modules:system",
				Arrays.asList("acoustic", "solid-borne", "vibration", "current"), Arrays.asList("mean", "variance")	,iiwowaSemaphore));
		*/
	}
	
	//##########################################################################################
	/**
	 * Generate features by checking for possible combinations of sensors and algorithms.
	 */
	public void initFeatures(){
		
		//Send reset message to intelligent sensors to destroy "old" features and their socket communication interfaces
		for(int i=0; i<listOfIntSens.size(); i++){
			sendClearMessageToIntSens(listOfIntSens.get(i).getIP());
		}
		
		//Sensors
		for(int i = 0; i < listOfSolidSens.size(); i++){
			
			int dimSens = listOfSolidSens.get(i).getOutputDimensonality();
			String typeSensSig = listOfSolidSens.get(i).getPhysPheno();
			
			//Algorithms
			for(int j = 0; j < listOfAlgorithms.size(); j++){
				
				int dimAlg = listOfAlgorithms.get(j).getInputDimensonality();
				String[] typeAlgSig = listOfAlgorithms.get(j).getPhysPheno();
//				System.out.println("Type Alg: "+ typeAlgSig + "Type Sens: " + typeSensSig);
//				System.out.println("Type Alg: "+ dimAlg + "Type Sens: " + dimSens);
//				System.out.println("---");
				
				if(dimAlg==dimSens){
					for(int k=0; k<typeAlgSig.length; k++){
						if(typeAlgSig[k].equals(typeSensSig))
						addFeature(listOfSolidSens.get(i), listOfAlgorithms.get(j));
					}
					
				}
			}			
		}
	}
	
	//##########################################################################################
	/**
	 * Assign features to physical, funtional and quality attributes.
	 */
	private void orchestrateGeneralAttributes(){
		for(int i = 0; i < attributeStore.getAttributeGeneralList().size(); i++){
			
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//Orchestrate functional and physical attributes
			if(attributeStore.getAttributeGeneralList().get(i).getAttrType().equals("functional") ||
					attributeStore.getAttributeGeneralList().get(i).getAttrType().equals("physical")){
				
				//Get foi of the attribute
				String foiA = attributeStore.getAttributeGeneralList().get(i).getFoI();
				//Get physical phenomenas that are suitable for the attribute
				List<String> physG = attributeStore.getAttributeGeneralList().get(i).getSuitablePhysPheno();
				List<String> featG = attributeStore.getAttributeGeneralList().get(i).getSuitableFeatures();
				
				//Get childs of the module (top-down)
				List<String> childs = new ArrayList<String>();
				childs.add(foiA); //Add the foi itself to the child list (simplifies the following iteration)
				getChildList(foiA, childs);
				
				//Run through features
				for(int j = 0; j < featureStore.getFeatureList().size(); j++){
					boolean fits = false;
					
					//Get foi of the feature
					String foiS = featureStore.getFeatureList().get(j).getfoi();
					
					//Get the identifier of the feature
					String featS = featureStore.getFeatureList().get(j).getAlgorithm().getUID();
					
					//Get physical phenomena of the feature
					String physS = featureStore.getFeatureList().get(j).getSuitablePhysPheno().get(0);
					
					//Run through childs of FoI
					for(int k=0; k<childs.size();k++){
						
						//Check if Attribute and feature have equal FoI
						if(foiS.equals(childs.get(k))){						
							
							boolean fitPhys = false;
							boolean fitFeat = false;
							//Check if the Sensor measures a physical phenomena suitable for the Attribute
							for(int l=0; l<physG.size(); l++){
								if(physG.get(l).equals(physS)){
									fitPhys = true;
								}
							}
							
							if(featG != null){
								for(int l=0; l<featG.size(); l++){
									if(featG.get(l).equals(featS)){
										fitFeat = true;
									}
								}
							}
							else fitFeat = true;
							
							if(fitPhys && fitFeat){
								fits=true;
							}
						}
					}
					//If a match was found add the feature (UID) to the attribute
					if(fits){
						attributeStore.getAttributeGeneralList().get(i).addFeature(featureStore.getFeatureList().get(j));
					}	
				}
			}
			
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			//Orchestrate quality/product attributes
			if(attributeStore.getAttributeGeneralList().get(i).getAttrType().equals("quality")){
				
				//Get foi of the attribute
				String foiA = attributeStore.getAttributeGeneralList().get(i).getFoI();
				
				//Get physical phenomena of the attribute
				List<String> physG = attributeStore.getAttributeGeneralList().get(i).getSuitablePhysPheno();

				//Run through features
				for(int j = 0; j < featureStore.getFeatureList().size(); j++){
					boolean fits = false;
					
					//Get foi of the feature
					String foiS = featureStore.getFeatureList().get(j).getfoi();
					
					//Get physical phenomena of the feature
					String physS = featureStore.getFeatureList().get(j).getSuitablePhysPheno().get(0);
					
					//Check if Attribute and Sensor have equal FoI
					if(foiS.equals(foiA)){
					
						//Check if the Sensor measures a physical phenomena suitable for the Attribute
						for(int l=0; l<physG.size(); l++){
							if(physG.get(l).equals(physS)){
								fits = true;
							}
						}
						//If a match was found add the feature (UID) to the attribute					
						if(fits){
							attributeStore.getAttributeGeneralList().get(i).addFeature(featureStore.getFeatureList().get(j));
						}	
					}					
				}
			}
		}
	}	
	//##########################################################################################
	/**
	 * Assign features to module attributes.
	 */
	private void orchestrateModuleAttributes(){
		//Module-Attributes
		for(int i = 0; i < attributeStore.getAttributeModuleList().size(); i++){
			
			String foiA = attributeStore.getAttributeModuleList().get(i).getFoi();
			
			//Run through features
			for(int j = 0; j < featureStore.getFeatureList().size(); j++){
				boolean fits = false;
				String foiS = featureStore.getFeatureList().get(j).getfoi();
				
				//Check if feature observes the module
				//There should be no module attribute for the overall system
				if(foiS.equals(foiA) && !foiS.equals("inIT:modules:system")){					
					fits=true;
				}
				
				if(fits){
					attributeStore.getAttributeModuleList().get(i).addFeature(featureStore.getFeatureList().get(j));
					
					addToUpperModuleAttribute(foiS, featureStore.getFeatureList().get(j));
				}
				//System.out.println(listOfSolidSens.get(j).getFeaturesOfInterest().getFeatureList().get(0).getIdentifier().getValue());
			}
		}
	}
	//##########################################################################################
	/**
	 * (i) Remove Attributes with less than two features.
	 * (ii) Check for Attributes with equal feature. Physical Attributes and Attributes with
	 * a module that is low in the hierarchy of modules are preferred.
	 */
	private void removeInsufficientAttributes(){
		
		// Remove attributes whose cardinality is < 2
		attributeStore.removeBadAttributes();
		
		//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		// Remove attributes with an equal set of features
		List<AttributeGeneral> generalAttrs = attributeStore.getAttributeGeneralList();
		List<AttributeModule> moduleAttrs = attributeStore.getAttributeModuleList();
		//System.out.println("Attributes: " + (attributeStore.getAttributeGeneralList().size() + attributeStore.getAttributeModuleList().size()));
		
		/*
		AttributeGeneral a= generalAttrs.get(0);
		AttributeGeneral b= generalAttrs.get(1);
		generalAttrs.set(0, b);
		generalAttrs.set(1, a);
		*/
		
		
		// Check general against general and module attributes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		for(int i=1; i <= generalAttrs.size(); i++){
			AttributeGeneral attr1 = generalAttrs.get(i-1);
			
			// Check general against general attributes
			for(int j=i; j < generalAttrs.size(); j++){
				AttributeGeneral attr2 = generalAttrs.get(j);
				if(attr1.getFeatures().equals(attr2.getFeatures())){
					List<String> parents = new ArrayList<String>();
					List<String> childs = new ArrayList<String>();
					getParentList(attr1.getFoI(), parents);
					getChildList(attr1.getFoI(), childs);
															
					if(parents.contains(attr2.getFoI())){
						generalAttrs.remove(generalAttrs.indexOf(attr2));
						--j;
					}
					else if(childs.contains(attr2.getFoI())){
						generalAttrs.remove(generalAttrs.indexOf(attr1));
						--i;
						break;
					}
				}
			}
			
			// Check general against module attributes
			for(int j=0; j < moduleAttrs.size(); j++){
				AttributeModule attr2 = moduleAttrs.get(j);
				//System.out.println(attr1.getFeaturesStringList() + "---" + attr2.getFeaturesStringList());
				if(attr1.getFeaturesStringList().equals(attr2.getFeaturesStringList())){
					moduleAttrs.remove(moduleAttrs.indexOf(attr2));
					--j;
				}
			}			
		}
		// Check module against module attributes %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		for(int i=1; i <= moduleAttrs.size(); i++){
			AttributeModule attr1 = moduleAttrs.get(i-1);
			
			// Check general against general attributes
			for(int j=i; j < moduleAttrs.size(); j++){
				AttributeModule attr2 = moduleAttrs.get(j);
				
				if(attr1.getFeatures().equals(attr2.getFeatures())){
					List<String> parents = new ArrayList<String>();
					List<String> childs = new ArrayList<String>();
					getParentList(attr1.getFoi(), parents);
					getChildList(attr1.getFoi(), childs);
					
					if(parents.contains(attr2.getFoi())){
						moduleAttrs.remove(moduleAttrs.indexOf(attr2));
						--j;
					}
					else if(childs.contains(attr2.getFoi())){
						moduleAttrs.remove(moduleAttrs.indexOf(attr1));
						--i;
						break;
					}
					else{
						generalAttrs.remove(generalAttrs.indexOf(attr1));
						--i;
						break;
					}
				}
			}
		}
	}
	//##########################################################################################
	/**
	 * Start threads for attributes. The threads open ports to receive input membership values coming
	 * from features (intelligent sensors).
	 */
	private void startAttributeThreads(){
		
		
		List<AttributeGeneral> generalAttrs = attributeStore.getAttributeGeneralList();
		List<AttributeModule> moduleAttrs = attributeStore.getAttributeModuleList();
		
		
		//Define attribute input ports and start threads that observe these ports
		int port = 9700;
		for(int i = 0; i<generalAttrs.size();i++){
			for(int j = 0; j < generalAttrs.get(i).getFeatures().size(); j++){
				generalAttrs.get(i).getFeatures().get(j).addAttributePort(port);
				generalAttrs.get(i).addInputPort(port);
				port++;
			}
			generalAttrs.get(i).initThreads();
		}
		for(int i = 0; i<moduleAttrs.size();i++){
			for(int j = 0; j < moduleAttrs.get(i).getFeatures().size(); j++){
				moduleAttrs.get(i).getFeatures().get(j).addAttributePort(port);
				moduleAttrs.get(i).addInputPort(port);
				port++;
			}
			moduleAttrs.get(i).initThreads();
		}
		attributeStore.initIIWOWA();
		attributeStore.startWriteCsvThread();		
	}
	//##########################################################################################
	/**
	 * Returns the uid's of child modules of the foi
	 * @param foi The UID of the module (foi)
	 * @param childs The list where child modules are written to.
	 */
	private void getChildList(String foi, List<String> childs){
		
		PhysicalSystem module = getModule(foi);
		for(int k = 0; k < module.getIdentificationList().get(0).getIdentifier2List().size(); k++){
			
			String label = module.getIdentificationList().get(0).getIdentifier2List().get(k).getLabel();
			String child = module.getIdentificationList().get(0).getIdentifier2List().get(k).getValue();
			
			if(label.equals("child") & !child.equals("bottom")){
				childs.add(child);
				getChildList(child, childs);
			}
		}
	}
	//##########################################################################################
	/**
	 * Returns the uid's of child modules of the foi
	 * @param foi The UID of the module (foi)
	 * @param parents The list where parent modules are written to.
	 */
	private void getParentList(String foi, List<String> parents){
		
		//System.out.println("FOI: " + foi + "");
		PhysicalSystem module = getModule(foi);
		
		//System.out.println(foi);
		for(int k = 0; k < module.getIdentificationList().get(0).getIdentifier2List().size(); k++){
			
			String label = module.getIdentificationList().get(0).getIdentifier2List().get(k).getLabel();
			String parent = module.getIdentificationList().get(0).getIdentifier2List().get(k).getValue();
			if(label.equals("parent") & !parent.equals("top")){
				parents.add(parent);
				getParentList(parent, parents);
			}
		}	
	}	
	//##########################################################################################
	/**
	 * Add a feature to the attribute of modules that are higher in the hierarchy of modules.
	 * The method is called recursively.
	 */
	private void addToUpperModuleAttribute(String foi, Feature feature){
		
		PhysicalSystem module = smlHelper.newPhysicalSystem();
		//Get the module the sensor is observing
		for(int l = 0; l < listOfModules.size(); l++){
			
			if(listOfModules.get(l).getIdentifier().getValue().equals(foi)){
				module = listOfModules.get(l);
				String upperModule = module.getIdentificationList().get(0).getIdentifier2List().get(0).getValue();				
				//System.out.println("Module: " + listOfModules.get(l).getIdentifier().getValue());
				
				//There should be no module attribute for the overall system
				if(!upperModule.equals("inIT:modules:system") && !upperModule.equals("top")){
					//System.out.println("upper: " +upperModule);
					attributeStore.getAttributeModule(upperModule).addFeature(feature);
					addToUpperModuleAttribute(upperModule, feature);
				}
			}
		}
	}	
	//############################################################################################
	/**
	 *  Add the physical quantities of sensors that are attached to a module to the module.
	 *  Required to automatically generate physical attributes for modules.
	 */
	private void addPhysQuantToModules(){
		
		//Run through sensors
		for(int i = 0; i < listOfSolidSens.size(); i++){
			
			//Get foi of the sensor
			//System.out.println(listOfSolidSens.get(i).getClassificationList().get(0).getClassifierList().get(1).getId());
			String foiS = listOfSolidSens.get(i).getfoi();
			
			if(!foiS.equals("inIT:product")){
				//Get childs of the module (top-down)
				List<String> parents = new ArrayList<String>();
				parents.add(foiS); //Add the foi itself to the child list (simplifies the following iteration)
				getParentList(foiS, parents);
				
				//Add the physical phenomena that the sensor measures to the module
				Term physPhen = smlHelper.newTerm();
				String physPheno = listOfSolidSens.get(i).getPhysPheno();
				physPhen.setValue(physPheno);	
				
				for(int k = 0; k<parents.size(); k++){
					//Initialize new classification section if it is not available yet
					if(getModule(parents.get(k)).getClassificationList().isEmpty()){
						ClassifierList clLisz = smlHelper.newClassifierList();
						getModule(parents.get(k)).addClassification(clLisz);
						getModule(parents.get(k)).getClassificationList().get(0).addClassifier(physPhen);
					}
					else{ //Check if Physical Quantity is already existing. if not add it
						boolean match = false;
						for(int l = 0; l<getModule(parents.get(k)).getClassificationList().get(0).getClassifierList().size(); l++){
							if( getModule(parents.get(k)).getClassificationList().get(0).getClassifierList().get(l).getValue().equals(physPhen.getValue())){
								match = true;
							}
						}
						if(!match){
							getModule(parents.get(k)).getClassificationList().get(0).addClassifier(physPhen);
						}
					}
				}
			}
		}
	}
	
	//##########################################################################################
	/**
	 *  Add a feature to the feature list (a combination of a sensor and a algorithm).
	 * @param sens The used sensor
	 * @param alg The algorithm for feature extraction
	 */
	private void addFeature(Sensor sens, Algorithm alg){
				
		//For numbering of features
		int numFeatures;
		if(featureStore.getFeatureList().isEmpty())
			numFeatures = 0;
		else
			numFeatures = featureStore.getFeatureList().size();

		
		List<String> physPhen = new ArrayList<String>();
		physPhen.add(sens.getPhysPheno());
				
		String foi = sens.getfoi();
		
		// Implementation is for single sensors currently
		List<String> sensors = new ArrayList<String>();
		sensors.add(sens.getUID());
		
		//Init feature object
		String[] sensName = sens.getUID().split(":");
		Feature feature = new Feature("inIT:features:" + alg.getUID() + ":" + sensName[2], physPhen, alg, foi);
		feature.addSensor(sens);
		
		//Send feature information to intelligent sensor
		String message = "feature_"+ alg.getUID() + "_" +feature.getUID();
		int ports = sendFeatureToIntSens(alg.getIP(), message);
		
		//Set ports for data exchange
		feature.setPorts(ports, ports+1);		
		
		//Add feature to the servers's feature store 
		featureStore.addFeature(feature);
	}
	//##########################################################################################
	/**
	 * Send information for configuration of features to intelligent sensor.
	 * @param message
	 */
	@SuppressWarnings("resource")
	private int sendFeatureToIntSens(String ip, String message){
		int ports = 0;		
		//get the localhost IP address, if server is running on some other IP, you need to use that
        InetAddress host;
		try {
			
			host = InetAddress.getByName(ip);
	        Socket socket = null;
	        ObjectOutputStream oos = null;
	        ObjectInputStream ois = null;
	        
	        long start = System.currentTimeMillis();
	        //establish socket connection to server
	        socket = new Socket(host.getHostName(), 9900);
	        //write to socket using ObjectOutputStream
	        oos = new ObjectOutputStream(socket.getOutputStream());
	        oos.writeObject(message);
	        //read the server response message
	        ois = new ObjectInputStream(socket.getInputStream());
	        ports = (int) ois.readObject();
	        System.out.println("Time to communicate feature information: " + (System.currentTimeMillis() - start) / 1000);
	        //System.out.println("Port: " + ports);
	        //close resources
	        ois.close();
	        oos.close();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}  
		
		return ports;
	}
	//##########################################################################################
	/**
	 * Send message to clear feature interfaces to intelligent sensor.
	 * @param message
	 */
	@SuppressWarnings("resource")
	private void sendClearMessageToIntSens(String ip){	
		//get the localhost IP address, if server is running on some other IP, you need to use that
        InetAddress host;
		try {
			host = InetAddress.getByName(ip);
	        Socket socket = null;
	        ObjectOutputStream oos = null;
	        ObjectInputStream ois = null;
	        //establish socket connection to server
	        socket = new Socket(host.getHostName(), 9900);
	        //write to socket using ObjectOutputStream
	        oos = new ObjectOutputStream(socket.getOutputStream());
	        oos.writeObject("clear");
	        //read the server response message
	        ois = new ObjectInputStream(socket.getInputStream());

	        //close resources
	        ois.close();
	        oos.close();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	//##########################################################################################
	/**
	 * Check if one some intelligent sensor with a specific UID already exists
	 * @param uid
	 * @return
	 */
	public boolean checkIntSensUID(String uid){
		boolean exists = false;
		
		for(int i=0; i < listOfIntSens.size();i++){
			
			if(listOfIntSens.get(i).getUID().equals(uid))
				exists=true;
		}
	
		return exists;
	}
	
	//##########################################################################################
	/**
	 * Write results to XML-File
	 */
	private void writeResults(){
		
		try {			
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
			
			// root elements
			Document doc = docBuilder.newDocument();
			Element root = doc.createElement("attributeList");
			doc.appendChild(root);			
			
			for(int i = 0; i < attributeStore.getAttributeGeneralList().size(); i++){
				Element attribute = doc.createElement("attribute");
				attribute.setAttribute("uid", attributeStore.getAttributeGeneralList().get(i).getUID());
				root.appendChild(attribute);
				
				Element type = doc.createElement("type");
				type.appendChild(doc.createTextNode(attributeStore.getAttributeGeneralList().get(i).getAttrType()));
				attribute.appendChild(type);
				
				Element foi = doc.createElement("foi");
				foi.appendChild(doc.createTextNode(attributeStore.getAttributeGeneralList().get(i).getFoI()));
				attribute.appendChild(foi);
				
				Element sources = doc.createElement("sourceList");
				attribute.appendChild(sources);
				
				for(int j = 0; j< attributeStore.getAttributeGeneralList().get(i).getSensors().size(); j++){
					Element source = doc.createElement("source");
					source.setAttribute("uid", attributeStore.getAttributeGeneralList().get(i).getSensors().get(j));
					sources.appendChild(source);
				}
			}
			
			for(int i = 0; i < attributeStore.getAttributeModuleList().size(); i++){
				Element attribute = doc.createElement("attribute");
				attribute.setAttribute("uid", attributeStore.getAttributeModuleList().get(i).getUID());
				root.appendChild(attribute);
				
				Element type = doc.createElement("type");
				type.appendChild(doc.createTextNode("module"));
				attribute.appendChild(type);
				
				Element foi = doc.createElement("foi");
				foi.appendChild(doc.createTextNode(attributeStore.getAttributeModuleList().get(i).getFoi()));
				attribute.appendChild(foi);
				
				Element sources = doc.createElement("sourceList");
				attribute.appendChild(sources);
				
				for(int j = 0; j< attributeStore.getAttributeModuleList().get(i).getSensors().size(); j++){
					Element source = doc.createElement("source");
					source.setAttribute("uid", attributeStore.getAttributeModuleList().get(i).getSensors().get(j));
					sources.appendChild(source);
				}
			}
			
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			//transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File("result.xml"));
			
			transformer.transform(source, result);			
		} catch (ParserConfigurationException | TransformerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}	
	//##########################################################################################
	/**
	 * Display results to console
	 */
	private void printResults(){
		
		System.out.println("\n############### Results ################ ");
		System.out.println("\nFeatures: " + featureStore.getFeatureList().size());
//		for(int i =0; i<listOfFeatures.size(); i++){
//			System.out.println(listOfFeatures.get(i).getIdentifier().getValue());
//			System.out.println(listOfFeatures.get(i).getClassificationList().get(0).getClassifierList().get(1).getValue()
//					+" ... "
//					+ listOfFeatures.get(i).getClassificationList().get(0).getClassifierList().get(3).getValue());
//			
//		}
		System.out.println("Modules: " + listOfModules.size());
		
		System.out.println("\n######### Available Attributes ######### ");
		System.out.println(String.format("%s - %-50s - %-25s - %-25s - %-50s",
				"#", "UID", "Attribute Type", "Module", "List of features"));
		int atrNum = 1;
		for(int i = 0; i < attributeStore.getAttributeGeneralList().size(); i++){
			System.out.println(String.format("%s - %-50s - %-25s - %-25s - %-50s",
					String.valueOf(atrNum),
					attributeStore.getAttributeGeneralList().get(i).getUID(),
					attributeStore.getAttributeGeneralList().get(i).getAttrType(),
					attributeStore.getAttributeGeneralList().get(i).getFoI(),
					attributeStore.getAttributeGeneralList().get(i).getFeaturesStringList()));
			atrNum++;
		}
		for(int i = 0; i < attributeStore.getAttributeModuleList().size(); i++){
			System.out.println(String.format("%s - %-50s - %-25s - %-25s - %-50s",
					String.valueOf(atrNum),
					attributeStore.getAttributeModuleList().get(i).getUID(),
					"module",
					attributeStore.getAttributeModuleList().get(i).getFoi(),
					attributeStore.getAttributeModuleList().get(i).getFeaturesStringList()));
			atrNum++;
		}
		System.out.print("\n\n### Press Enter to end the programm ###\n");
	}	
	//##########################################################################################
	//##########################################################################################
	//################################ add get remove  #########################################
	//Setter
	public void addIntelligenSensor(IntelligentSensor sensor){
		boolean newSensor = true;
		//Check if the intelligent sensor is already known
		for(int i = 0; i<listOfIntSens.size();i++){
			if(listOfSolidSens.get(i).getUID().equals(sensor.getUID())){
				newSensor = false;
			}	
		}
		//Add intelligent sensor if it is new
		if(newSensor) listOfIntSens.add(sensor);
	}
	
	//##########################################################################################
	/**
	 * The intelligent sensor and all of its components (sensors and algorithms) are removed 
	 * from the Knowledge Base.
	 * @param uid The identifier of the intelligent sensor
	 */
	public void removeIntelligenSensor(String uid){
		for(int i = 0; i < listOfIntSens.size(); i++){
			//System.out.print(listOfIntSens.get(i).getUniqueIdentifier() + "   ...");
			if(listOfIntSens.get(i).getUID().equals(uid)){				
				// remove solid sensors that are available from the intelligent sensor
				for(int j=0; j<listOfIntSens.get(i).getElementarySensors().size();j++){
					removeSolidSensor(listOfIntSens.get(i).getElementarySensors().get(j).getUID());
				}
				for(int j=0; j<listOfIntSens.get(i).getAlgorithms().size();j++){
					removeAlgorithm(listOfIntSens.get(i).getAlgorithms().get(j).getUID());
				}				
				// remove the intelligent sensor from the list of available intelligent sensors
				System.out.println("Intelligent Sensor " + uid + " and its components are removed from the KB.");
				listOfIntSens.remove(i);
			}
		}
	}
	//##########################################################################################
	public void addSolidSensor(Sensor sensor){
		listOfSolidSens.add(sensor);
	}	
	//##########################################################################################
	public void removeSolidSensor(String uid){
		
		for(int i = 0; i < listOfSolidSens.size(); i++){
			if(listOfSolidSens.get(i).getUID().equals(uid)){
				//System.out.println("Sensor " + uid + " removed.");
				listOfSolidSens.remove(i);
			}
		}
	}
	//##########################################################################################
	public void addAlgorithm(Algorithm algorithm){
		listOfAlgorithms.add(algorithm);
	}
	//##########################################################################################
	public void removeAlgorithm(String uid){
		
		for(int i = 0; i < listOfAlgorithms.size(); i++){
			if(listOfAlgorithms.get(i).getUID().equals(uid)){
				//System.out.println("Algorithm " + uid + " removed.");
				listOfAlgorithms.remove(i);
			}
		}
	}
	//##########################################################################################
	public void addModule(String module){
		int numModules = listOfModules.size();
		PhysicalSystem mod;
		boolean match = false;
		try {
			mod = (PhysicalSystem) smlUtils.readProcess(new FileInputStream(new File(module)));	
			if(numModules == 0){
				listOfModules.add(mod);
			}
			else{
				for(int i = 0; i<numModules;i++){
					if(listOfModules.get(i).getIdentifier().toString().equals(mod.getIdentifier().toString())){
						match = true;
					}
				}
				if(!match){
					listOfModules.add(mod);
				}
			}
		} catch (XMLReaderException | FileNotFoundException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
	}
	//##########################################################################################
	//##########################################################################################
	
	//Getter
	public List<IntelligentSensor> getIntelligentSensorList(){
		
		if(listOfIntSens.size() > 0) return listOfIntSens;
		else return null;
	}
	//##########################################################################################
	public List<Feature> getFeatureList(){
		
		if(featureStore.getFeatureList().size() > 0) return featureStore.getFeatureList();
		else return null;
	}
	//##########################################################################################
	public List<Algorithm> getAlgorithmList(){
		
		if(listOfAlgorithms.size() > 0) return listOfAlgorithms;
		else return null;
	}
	//##########################################################################################
	public List<PhysicalSystem> getModuleList(){
			
		if(listOfModules.size() > 0) return listOfModules;
		else return null;
	}
	//##########################################################################################
	public IntelligentSensor getIntelligentSensor(int index){
		
		if(listOfIntSens.size() > 0) return listOfIntSens.get(index);
		else return null;
	}
	//##########################################################################################
	public IntelligentSensor getIntelligentSensor(String uid){
		
		if(listOfIntSens.size() > 0){
			for(int i = 0; i< listOfIntSens.size();i++){
				if(listOfIntSens.get(i).getUID().equals(uid))
					return listOfIntSens.get(i);
			}
		}
		return null;
	}
	//##########################################################################################
	public Sensor getSolidSensor(int index){
			
		if(listOfSolidSens.size() > 0) return listOfSolidSens.get(index);
		else return null;
	}
	//##########################################################################################
	public List<Sensor> getSolidSensorList(){
			
		if(listOfSolidSens.size() > 0) return listOfSolidSens;
		else return null;
	}
	//##########################################################################################
	public Sensor getSolidSensor(String UID){
		
		for(int i = 0; i< listOfSolidSens.size(); i++){
			
			if(listOfSolidSens.get(i).getUID().equals(UID))
				return listOfSolidSens.get(i);			
		}
		return null;
	}
	//##########################################################################################
	public Algorithm getAlgorithm(int index){
		
		if(listOfAlgorithms.size() > 0) return listOfAlgorithms.get(index);
		else return null;
	}
	//##########################################################################################
	public Algorithm getAlgorithm(String UID){
		
		for(int i = 0; i< listOfAlgorithms.size(); i++){
			
			if(listOfAlgorithms.get(i).getUID().equals(UID))
				return listOfAlgorithms.get(i);			
		}
		return null;
	}
	//##########################################################################################
	public PhysicalSystem getModule(int index){
			
		if(listOfModules.size() > 0) return listOfModules.get(index);
		else return null;
	}
	//##########################################################################################
	public PhysicalSystem getModule(String UID){
		
		for(int i = 0; i< listOfModules.size(); i++){
			
			if(listOfModules.get(i).getIdentifier().getValue().equals(UID))
				return listOfModules.get(i);			
		}
		return null;
	}
	//##########################################################################################
	public Term getSensorIdentifier(PhysicalComponent sensor, String name){
		
		for(int i = 0; i< sensor.getIdentificationList().get(0).getIdentifier2List().size(); i++){
			
			if(sensor.getIdentificationList().get(0).getIdentifier2List().get(i).getLabel().equals(name))
				return sensor.getIdentificationList().get(0).getIdentifier2List().get(i);			
		}
		return null;
	}
	//##########################################################################################
	public Term getClassifier(PhysicalComponent sensor, String name){
		for(int i = 0; i< sensor.getClassificationList().get(0).getClassifierList().size(); i++){
			
			if(sensor.getClassificationList().get(0).getClassifierList().get(i).getLabel().equals(name))
				return sensor.getClassificationList().get(0).getClassifierList().get(i);			
		}
		return null;
	}
	
	/**
	 * Write semantics to an XML-File for Profinet configurations
	 * IMPORTANT: profinet-inputs (data of this application -> Profinet)
	 *			  profinet-outputs (data from Profinet -> this application)
	 */
	public void generateProfinetXML(){
		try {
			
			//UUID
			UUID id = UUID.randomUUID();
			
			//IDs for features and sensors. Required for UI
			int IDF = 0;
			int IDS = 0;
			
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
			// root elements
			Document doc = docBuilder.newDocument();
			Element root = doc.createElement("ProfinetConfig");
			root.setAttribute("uuid", id.toString());
			doc.appendChild(root);
			
			// System Manager elements ###########################################################################
			Element sysmanager = doc.createElement("DevPara");
			sysmanager.setAttribute("stationName", "sysmanager");
			root.appendChild(sysmanager);
			
			//ip
			Element addrSysManager = doc.createElement("IP");
			String[] addrSplit = SysManagerIP.split("\\.");
			addrSysManager.setAttribute("d1", addrSplit[0]);
			addrSysManager.setAttribute("d2", addrSplit[1]);
			addrSysManager.setAttribute("d3", addrSplit[2]);
			addrSysManager.setAttribute("d4", addrSplit[3]);
			sysmanager.appendChild(addrSysManager);
			
			//mask
			Element maskSysManager = doc.createElement("Mask");
			maskSysManager.setAttribute("d1", "255");
			maskSysManager.setAttribute("d2", "255");
			maskSysManager.setAttribute("d3", "255");
			maskSysManager.setAttribute("d4", "0");
			sysmanager.appendChild(maskSysManager);
			
			Element modulesSysManager = doc.createElement("Modules");
			sysmanager.appendChild(modulesSysManager);
			
			//Inputs for Attributes
			for(int i = 0; i<attributeStore.getAttributeGeneralList().size(); i++){
				//Incoming feature (membership) values for attributes
				for(int j = 0; j<attributeStore.getAttributeGeneralList().get(i).getInputPorts().size();j++){
					Element module = doc.createElement("Module");
					module.setAttribute("type", "output");
					module.setAttribute("datalength", String.valueOf(8)); //Double
					module.setAttribute("port", String.valueOf(attributeStore.getAttributeGeneralList().get(i).getInputPorts().get(j)));
					modulesSysManager.appendChild(module);
				}				
			}
			for(int i = 0; i<attributeStore.getAttributeModuleList().size(); i++){
				//Incoming feature (membership) values for attributes
				for(int j = 0; j<attributeStore.getAttributeModuleList().get(i).getInputPorts().size();j++){
					Element module = doc.createElement("Module");
					module.setAttribute("type", "output");
					module.setAttribute("datalength", String.valueOf(8)); //Double
					module.setAttribute("port", String.valueOf(attributeStore.getAttributeModuleList().get(i).getInputPorts().get(j)));
					modulesSysManager.appendChild(module);
				}				
			}
			
			// intelligent sensor elements	######################################################################		
			for(int j=0; j<listOfIntSens.size();j++){
				Element intSens = doc.createElement("DevPara");
				intSens.setAttribute("stationName", listOfIntSens.get(j).getUID());
				root.appendChild(intSens);
				
				Element addr = doc.createElement("IP");
				String dummy = listOfIntSens.get(j).getIP();
				String[] addrS = dummy.split("\\.");
				addr.setAttribute("d1", addrS[0]);
				addr.setAttribute("d2", addrS[1]);
				addr.setAttribute("d3", addrS[2]);
				addr.setAttribute("d4", addrS[3]);
				intSens.appendChild(addr);
				
				Element mask = doc.createElement("Mask");
				mask.setAttribute("d1", "255");
				mask.setAttribute("d2", "255");
				mask.setAttribute("d3", "255");
				mask.setAttribute("d4", "0");
				intSens.appendChild(mask);
				
				Element modules = doc.createElement("Modules");
				intSens.appendChild(modules);
				
				//Run through available elementary sensors of an intelligent sensor
				for(int i = 0; i<listOfIntSens.get(j).getElementarySensors().size();i++){
					Sensor sensor = listOfIntSens.get(j).getElementarySensors().get(i);
					Element module = doc.createElement("Module");
					module.setAttribute("type", "input");
					module.setAttribute("datalength", String.valueOf(sensor.getDataLength()));
					module.setAttribute("port", String.valueOf(sensor.getPort()));
					module.setAttribute("class", "Sensor");
					module.setAttribute("id", "S" + String.valueOf(IDS));
					module.setAttribute("name", listOfIntSens.get(j).getElementarySensors().get(i).getUID());
					// increment sensor id
					IDS++;
					modules.appendChild(module);
					
					//Get Features that require the sensor's measurements for feature computation
					//The feature has some address where sensor data has to be send to
					List<Feature> features = featureStore.getFeaturesWithSensor(sensor.getUID());
					for(int k = 0; k<features.size(); k++){
						Element dest = doc.createElement("Destination");
						String[] ip = features.get(k).getAlgorithm().getIP().split("\\.");
						dest.setAttribute("d1", ip[0]);
						dest.setAttribute("d2", ip[1]);
						dest.setAttribute("d3", ip[2]);
						dest.setAttribute("d4", ip[3]);						
						dest.setAttribute("desPort", String.valueOf(features.get(k).getInPort()));
						module.appendChild(dest);
					}
				}
				
				//Algorithms are not only profinet-inputs (data of this application -> Profinet).
				//They are profinet-outputs (data from Profinet -> this application) as well.
				for(int i = 0; i<listOfIntSens.get(j).getAlgorithms().size(); i++){
					
					Algorithm alg = listOfIntSens.get(j).getAlgorithms().get(i);
					
					//Features that incorporate an algorithm that is available from this intelligent sensor
					//have some input that is a profinet-output.
					List<Feature> features = featureStore.getFeaturesWithAlgorithm(alg.getUID());
					for(int k = 0; k<features.size(); k++){
						//Incoming sensor data for features
						Element module = doc.createElement("Module");
						module.setAttribute("type", "output");
						module.setAttribute("datalength", String.valueOf(alg.getDataLength()));
						module.setAttribute("port", String.valueOf(features.get(k).getInPort()));
						modules.appendChild(module);
						
						
						//Destination (attribute) address and port for feature values.
						Element module1 = doc.createElement("Module");
						module1.setAttribute("type", "input");
						module1.setAttribute("datalength", String.valueOf(8));
						module1.setAttribute("port", String.valueOf(features.get(k).getOutPort()));
						module1.setAttribute("class", "Feature");
						module1.setAttribute("id", "F" + String.valueOf(IDF));
						module1.setAttribute("name", features.get(k).getUID());
						//increment feature id
						IDF++;
						modules.appendChild(module1);
						for(int l = 0; l<features.get(k).getAttributePorts().size(); l++){
							Element dest = doc.createElement("Destination");
							String[] ip = SysManagerIP.split("\\.");
							dest.setAttribute("d1", ip[0]);
							dest.setAttribute("d2", ip[1]);
							dest.setAttribute("d3", ip[2]);
							dest.setAttribute("d4", ip[3]);				
							dest.setAttribute("desPort", String.valueOf(features.get(k).getAttributePorts().get(l)));
							module1.appendChild(dest);
						}
					}
				}
			}
			
			//Save XML-file
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.METHOD, "xml");
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			//transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File("/home/pi/Profinet/pnConfig.xml"));
			
			transformer.transform(source, result);
			
		}catch(Exception e){
			System.out.println("Exception in XML-writing: " + e.getMessage());
		}
	}
	public void close(){
		attributeStore.closeAttributeThreads();
		attributeStore.stoptWriteCsvThread();
	}
}
