/*
 * Copyright 2017 inIT - Institute Industrial IT

	This file is part of "Automated Fusion System Design and Adaptation Implementation.

    "Automated Fusion System Design and Adaptation Implementation is free software: 
	you can redistribute it and/or modify it under the terms of the GNU General Public 
	License as published by the Free Software Foundation, either version 3 of the 
	License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY 
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this 
	program.  If not, see <http://www.gnu.org/licenses/>.
	
 * A copy of the GNU General Public License version 3 is provided in the file LICENSE.gpl3.0.txt .
 */
package opcua.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Timer;
import java.util.TimerTask;

import org.opcfoundation.ua.builtintypes.LocalizedText;
import org.opcfoundation.ua.builtintypes.NodeId;
import org.opcfoundation.ua.builtintypes.UnsignedInteger;
import org.opcfoundation.ua.core.ApplicationDescription;
import org.opcfoundation.ua.core.ApplicationType;
import org.opcfoundation.ua.core.ServerState;
import org.opcfoundation.ua.core.ServerStatusDataType;
import org.opcfoundation.ua.transport.security.SecurityMode;
import org.vast.sensorML.SMLHelper;
import org.vast.swe.SWEHelper;

import com.prosysopc.ua.ApplicationIdentity;
import com.prosysopc.ua.SecureIdentityException;
import com.prosysopc.ua.client.ServerList;
import com.prosysopc.ua.client.ServerListException;
import com.prosysopc.ua.client.ServerStatusListener;
import com.prosysopc.ua.client.UaClient;
import fusion.manager.Algorithm;
import fusion.manager.IntelligentSensor;
import fusion.manager.Sensor;
import fusion.orchestration.KnowledgeBase;
import net.opengis.gml.v32.impl.GMLFactory;

/**
 * The main class for an OPC UA client that gathers information from intelligent
 * sensors and stores it into a knowledge base.
 * 
 * @author A_Fritze
 *
 */
public class ClientMain implements ServerStatusListener {

	// SML Factory
	SMLHelper smlHelper;
	SWEHelper sweHelper;
	GMLFactory gmlHelper;

	// The Knowledge Base holds available information for orchestration
	KnowledgeBase KB;
	
	//Folder name of describtion files
	String nameDescriptionFolder;

	// For each server an own client object is required. Therefore a List is
	// used
	ArrayList<UaClient> clients;

	// List of available servers returned by the discovery Server
	ServerList serverList;
	// Update time that specifies the interval the LDS is checked for new
	// servers
	final long updateTimeLDS = 3000;

	public ClientMain(String folderNameDescriptionFiles) {
		smlHelper = new SMLHelper();
		sweHelper = new SWEHelper();
		gmlHelper = new GMLFactory();
		nameDescriptionFolder = folderNameDescriptionFiles;
		KB = new KnowledgeBase(getIP(), folderNameDescriptionFiles);
		clients = new ArrayList<UaClient>();
	}

	// *******************************************************************************************************************
	// *******************************************************************************************************************
	/**
	 * Start the OPC UA client. Available servers are discovered from the
	 * discovery server that is available at the IP of discoveryServerUri. After
	 * all information are transferred from the intelligent senors (servers) to
	 * the client, automated orchestration is carried out.
	 * 
	 * @param discoveryServerUri
	 * @throws Exception
	 */
	public void start(final String discoveryServerUri) throws Exception {
		long startTime = System.currentTimeMillis();
		
		// Discover a new server list from a discovery server at URI
		try {
			serverList = new ServerList(discoveryServerUri);
		} catch (ServerListException e) {
			System.out.println("The discovery server ist not available." + e.getMessage());
			return;
		}

		// Print available servers
		System.out.println(String.format("%s - %-25s - %-15s - %-50s - %-50s - %s", "#", "Name", "Type", "Product",
				"Application", "Address"));
		for (int i = 0; i < serverList.size(); i++) {
			final ApplicationDescription s = serverList.get(i);
			System.out.println(String.format("%d - %-25s - %-15s - %-50s - %-50s - %s", i,
					s.getApplicationName().getText(), s.getApplicationType(), s.getProductUri(),
					s.getApplicationUri(), s.getDiscoveryUrls()[0]));
		}

		// Connect to available servers
		for (int i = 0; i < serverList.size(); i++) {
			// List of available servers includes the discovery server
			final ApplicationDescription s = serverList.get(i);

			// Connect to available Intelligent Sensors and read required
			// information
			if (s.getApplicationType().toString().equals("Server")
					&& s.getApplicationName().getText().toString().equals("IntelligentSensor")) {
				String uidISens = s.getProductUri();
				// check if some intelligent sensor with the same UID
				// already is available
				boolean exists = KB.checkIntSensUID(uidISens);
				if (!exists) {
					String[] parts = s.getApplicationUri().split(":");
					String ip = parts[1];
					clients.add(new UaClient("opc.tcp://" + ip + ":52520/OPCUA/IntelligentSensor"));
					clients.get(clients.size() - 1).setSecurityMode(SecurityMode.NONE);
					this.initialize(clients.get(clients.size() - 1), s.getProductUri());
					clients.get(clients.size() - 1).addServerStatusListener(this);
					clients.get(clients.size() - 1).connect();
					this.readIntSensInfo(clients.get(clients.size() - 1), ip, this.KB);
				} else {
					System.out.println("\nIntelligent Sensor (" + uidISens
							+ ") already exists! It is not added to the Knowledge Base");
				}
			}
		}
		// Run the orchestration procedure
		this.runOrchestration();

		// The timer registers changes in the list of servers of the local
		// discovery server
		Timer timer = new Timer();
		timer.schedule(new TimerTask() {

			@Override
			public void run() {
				try {
					checkLDS(discoveryServerUri);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}, 0, updateTimeLDS);

		long endTime = System.currentTimeMillis();

		long elapsed = (endTime - startTime) / 1000;
		System.out.println("Elapsed Time: " + elapsed + "s");

		// Wait for ENTER
		InputStreamReader isr = new InputStreamReader(System.in);
		BufferedReader br = new BufferedReader(isr);
		br.readLine();

		timer.cancel();

		// Disconnect from all intelligent sensors before the application is
		// closed
		System.out.print("# DISCONNECTING FROM SERVERS #\n");
		for (int i = 0; i < clients.size(); i++) {
			clients.get(i).disconnect();
		}
		KB.close();
		System.out.print("# FINISHED #\n");
	}

	// *******************************************************************************************************************
	/**
	 * Returns the IP of the local device
	 * @return
	 */
	private static String getIP() {
		Enumeration<?> en;
		try {
			en = NetworkInterface.getNetworkInterfaces();
			while (en.hasMoreElements()) {
				NetworkInterface n = (NetworkInterface) en.nextElement();
				Enumeration<?> ee = n.getInetAddresses();
				while (ee.hasMoreElements()) {
					InetAddress i = (InetAddress) ee.nextElement();
			        if(!i.getHostAddress().contains(":") && //no IPv6
			        		!i.getHostAddress().substring(0, Math.min(i.getHostAddress().length(), 9)).equals("127.0.0.1")){ //no localhost
						return i.getHostAddress();
					}
				}
			}
		} catch (SocketException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	// *******************************************************************************************************************
	/**
	 * Define a minimal ApplicationIdentity. If you use secure connections, you
	 * will also need to define the application instance certificate and manage
	 * server certificates. See the SampleConsoleClient.initialize() for a full
	 * example of that.
	 */
	protected void initialize(UaClient client, String productUID)
			throws SecureIdentityException, IOException, UnknownHostException {

		// *** Application Description is sent to the server
		ApplicationDescription appDescription = new ApplicationDescription();
		appDescription.setApplicationName(new LocalizedText("SimpleClient", Locale.ENGLISH));
		// 'localhost' (all lower case) in the URI is converted to the actual
		// host name of the computer in which the application is run
		appDescription.setApplicationUri("urn:localhost:UA:SimpleClient");
		appDescription.setProductUri(productUID);
		appDescription.setApplicationType(ApplicationType.Client);

		final ApplicationIdentity identity = new ApplicationIdentity();
		identity.setApplicationDescription(appDescription);
		client.setApplicationIdentity(identity);

	}

	// *******************************************************************************************************************
	/**
	 * Reads information from intelligent sensor's information model and stores it in the KB
	 * @param client
	 * @param ip
	 * @param KB
	 * @throws Exception
	 */
	protected void readIntSensInfo(UaClient client, String ip, KnowledgeBase KB) throws Exception {
		// int ns =
		// client.getNamespaceTable().getIndex("http:/init-owl.de/IntelligentSensor/");
		// NodeId myIntSensNodeID = new NodeId(3, "IntelligentSensor");
		// IntelligentSensor_Type myIntSens = (IntelligentSensor_Type)
		// client.getAddressSpace().getNode(myIntSensNodeID);
		// System.out.println(myIntSens.getNumAlgorithms());
		// client.getAddressSpace().getNode(Ids.IntelligentSensor_Type);
		// System.out.println(client.getAddressSpace().getNode(Ids.Algorithm_Type).getAttributes().getDisplayName());
		// IntelligentSensor_TypeImpl test = (IntelligentSensor_TypeImpl)
		// client.getAddressSpace().getNode(Ids.IntelligentSensor_Type);

		// UID
		// ---------------------------------------------------------------------------------------------
		String uidISens = client.readAttribute(new NodeId(3, "IntelligentSensor"), UnsignedInteger.getFromBits(3))
				.getValue().toString();

		// Number of sensors
		int numSens = client
				.readAttribute(new NodeId(3, "IntelligentSensor/2:NumSensors"), UnsignedInteger.getFromBits(13))
				.getValue().intValue();
		// Number of algorithms
		int numAlgs = client
				.readAttribute(new NodeId(3, "IntelligentSensor/2:NumAlgorithms"), UnsignedInteger.getFromBits(13))
				.getValue().intValue();

		System.out.println("\nIntelligent Sensor (" + client.getUri() + "):\n   Sensors: " + numSens
				+ "\n   Algorithms: " + numAlgs + "\n--------");

		IntelligentSensor iSens = new IntelligentSensor(uidISens, ip);

		// Add single sensors
		// ..................................................................................
		for (int i = 0; i < numSens; i++) {
			// UID
			// ---------------------------------------------------------------------------------------------
			String uidSens = client.readAttribute(new NodeId(3, "Sens" + i), UnsignedInteger.getFromBits(3)).getValue()
					.toString();

			// Physical Quantity
			String physPheno = client.readAttribute(new NodeId(3, "Sens" + i + "/2:Classifier/2:PhysicalQuantity"),
					UnsignedInteger.getFromBits(13)).getValue().toString();

			// FOI
			String foiComplete = client
					.readAttribute(new NodeId(3, "Sens" + i + "/2:Classifier/2:FOI"), UnsignedInteger.getFromBits(13))
					.getValue().toString();
			String foi = foiComplete.substring(1, foiComplete.length() - 1);
			String[] a = foi.split(", ");
			// String title = a[0];
			// String href = a[1];

			String observedEntity = a[0];

			// KB.addModule(href);

			// Output
			// ------------------------------------------------------------------------------------------
			int dimensionality = client
					.readAttribute(new NodeId(3, "Sens" + i + "/2:Output/2:Dimension"), UnsignedInteger.getFromBits(13))
					.getValue().intValue();
			int dataLength = client.readAttribute(new NodeId(3, "Sens" + i + "/2:Output/2:DataLength"),
					UnsignedInteger.getFromBits(13)).getValue().intValue();
			String dataType = client
					.readAttribute(new NodeId(3, "Sens" + i + "/2:Output/2:DataType"), UnsignedInteger.getFromBits(13))
					.getValue().toString();
			int port = client
					.readAttribute(new NodeId(3, "Sens" + i + "/2:Output/2:Port"), UnsignedInteger.getFromBits(13))
					.getValue().intValue();
			// Generate sensor object
			Sensor elementarySensor = new Sensor(uidSens, physPheno, dimensionality, dataType, dataLength, ip, port,
					observedEntity);

			// Add single sensor to Intelligent Sensor Object
			iSens.addElementarySensor(elementarySensor);

			// Add single sensor to Knowledge Base
			KB.addSolidSensor(elementarySensor);
		}

		// Add single algorithms
		// ...............................................................................
		for (int i = 0; i < numAlgs; i++) {

			// UID
			// ---------------------------------------------------------------------------------------------
			String uidAlg = client.readAttribute(new NodeId(3, "Algorithm" + i), UnsignedInteger.getFromBits(3))
					.getValue().toString();

			// Input
			// -------------------------------------------------------------------------------------------
			// System.out.println(client.readAttribute(new NodeId(3,
			// "Algorithm"+ i +"/2:Input/2:PhysicalQuantities"),
			// UnsignedInteger.getFromBits(13)).getValue().toString());
			String inputPhysPheno[] = client
					.readAttribute(new NodeId(3, "Algorithm" + i + "/2:Input/2:PhysicalQuantities"),
							UnsignedInteger.getFromBits(13))
					.getValue().toString().split(", ");
			inputPhysPheno[0] = inputPhysPheno[0].substring(1, inputPhysPheno[0].length());
			inputPhysPheno[inputPhysPheno.length - 1] = inputPhysPheno[inputPhysPheno.length - 1].substring(0,
					inputPhysPheno[inputPhysPheno.length - 1].length() - 1);
			int inDimensionality = client.readAttribute(new NodeId(3, "Algorithm" + i + "/2:Input/2:Dimension"),
					UnsignedInteger.getFromBits(13)).getValue().intValue();

			// Output
			// ------------------------------------------------------------------------------------------
			int outDimensionality = client.readAttribute(new NodeId(3, "Algorithm" + i + "/2:Output/2:Dimension"),
					UnsignedInteger.getFromBits(13)).getValue().intValue();

			// Generate algorithm object
			Algorithm alg = new Algorithm(uidAlg, inputPhysPheno, inDimensionality, outDimensionality, "double", 8, ip);

			// Add algorithm to Intelligent Sensor object
			iSens.addAlgorithm(alg);

			// Add algorithm to Knowledge Base
			KB.addAlgorithm(alg);
		}
		// Add intelligent sensor to Knowledge Base
		KB.addIntelligenSensor(iSens);

	}

	// *******************************************************************************************************************
	/**
	 * Execute SEFU/IFU system orchestration
	 */
	private void runOrchestration() {

		//reset features, attributes, and modules of KB
		KB.clearLists();

		// Available modules as well as their structure (child and top modules)
		// have to be available from the KB
		// -- Here they are modeled as SensorML components and are added
		// manually
		
		/*
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/motor1.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/motor2.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/motor3.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/product.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/roller1.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/roller2.xml");
		KB.addModule("System/"+ nameDescribtionFolder +"/Modules/system.xml");
		*/
		File path = new File("System/"+ nameDescriptionFolder +"/Modules");

	    File [] files = path.listFiles();
	    for (int i = 0; i < files.length; i++){
	        if (files[i].isFile()){ //this line weeds out other directories/folders
	        	KB.addModule(files[i].getPath());
	        }
	    }
	    
		//KB.addModule("System/"+ nameDescribtionFolder +"/Modules/room.xml");
		//KB.addModule("System/"+ nameDescribtionFolder +"/Modules/window.xml");

		// Run the orchestration procedure
		KB.initFeatures();
		KB.initAttributes();
		KB.orchestrate();

		// Generate the Profinet-XML
		KB.generateProfinetXML();
	}

	// *******************************************************************************************************************
	/**
	 * Monitor the LDS to detect attachments and detachments of intelligent
	 * sensors.
	 * 
	 * @param discoveryServerURL
	 * @throws Exception
	 */
	private void checkLDS(String discoveryServerURL) throws Exception {

		// Discover a new server list from a discovery server at URI
		ServerList currentServerList = new ServerList(discoveryServerURL);

		// System.out.println("LDS: " + currentServerList.size() + "..." +
		// serverList.size());
		// System.out.println(clients.size());

		if (currentServerList.size() > serverList.size()) {

			// if a new intelligent sensor is found a re-orchestration is
			// required
			boolean reorchestrate = false;

			// Check if new intelligent sensor is attached
			// start at i = 1 because 0 is the LDS itself
			for (int i = 1; i < currentServerList.size(); i++) {
				
				boolean newServer = true;
				final ApplicationDescription s = currentServerList.get(i);
				String[] parts = s.getApplicationUri().split(":");
				String ip = parts[1];

				// compare one server of currentServerList with all servers of
				// serverList
				for (int j = 1; j < serverList.size(); j++) {
					// if a server with equal applicationUri is available the
					// server is not new
					if (currentServerList.get(i).getProductUri().equals(serverList.get(j).getProductUri())) {
						newServer = false;
					}
				}
				if (KB.getIntelligentSensorList() != null) {
					if (KB.getIntelligentSensorList().size() > 0) {
						for (int j = 0; j < KB.getIntelligentSensorList().size(); j++) {
							if (KB.getIntelligentSensorList().get(j).getIP().equals(ip)) {
								newServer = false;
							}
						}
					}
				}

				// if intelligent sensor (server) is new add it to the system
				if (newServer) {

					// Connect to available Intelligent Sensors and read
					// required information
					if (s.getApplicationType().toString().equals("Server")
							&& s.getApplicationName().getText().toString().equals("IntelligentSensor")) {
						clients.add(new UaClient("opc.tcp://" + ip + ":52520/OPCUA/IntelligentSensor"));
						clients.get(clients.size() - 1).setSecurityMode(SecurityMode.NONE);
						initialize(clients.get(clients.size() - 1), s.getProductUri());
						clients.get(clients.size() - 1).addServerStatusListener(this);
						clients.get(clients.size() - 1).connect();
						clients.get(clients.size() - 1).registerModel(opcua.types.client.InformationModel.MODEL);
						// loadInformationModels(clients.get(clients.size()-1));
						readIntSensInfo(clients.get(clients.size() - 1), ip, KB);

						reorchestrate = true;

						System.out.println("\nNew intelligent sensor (" + s.getProductUri() + ")");
					}
				}
			}
			serverList = currentServerList;

			//In case of new intelligent sensor carry out re-orchestration
			if (reorchestrate) {
				runOrchestration();
			} else {
				System.out.println("\nNew intelligent sensor: UID already exists");
			}
		} else if (currentServerList.size() < serverList.size()) {
			System.out.println("\nLost intelligent sensor.");
			serverList = currentServerList;
		}
	}

	// *******************************************************************************************************************
	protected void loadInformationModels(UaClient client) {
		// Uncomment to take the extra information models in use.

		client.registerModel(opcua.types.client.InformationModel.MODEL);
		try {
			client.getAddressSpace()
					.loadModel(new File(System.getProperty("user.dir"), "codegen/intelligentsensor.xml").toURI());
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		// // Register generated classes
		client.registerModel(com.prosysopc.ua.types.di.client.InformationModel.MODEL);
		// server.registerModel(com.prosysopc.ua.types.adi.server.InformationModel.MODEL);
		// server.registerModel(com.prosysopc.ua.types.plc.server.InformationModel.MODEL);
		//
		// Load the standard information models
		try {
			client.getAddressSpace().loadModel(UaClient.class.getResource("Opc.Ua.Di.NodeSet2.xml").toURI());
			// server.getAddressSpace().loadModel(
			// UaServer.class.getResource("Opc.Ua.Adi.NodeSet2.xml")
			// .toURI());
			// server.getAddressSpace().loadModel(
			// UaServer.class.getResource("Opc.Ua.Plc.NodeSet2.xml")
			// .toURI());
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	// *******************************************************************************************************************
	/**
	 * Removes an intelligent sensor fro the list of available devices
	 * @param uaClient
	 */
	private void removeClient(UaClient uaClient) {

		int index = clients.indexOf(uaClient);
		clients.get(index).disconnect();
		clients.remove(index);
	}

	// *******************************************************************************************************************
	// Server Listener
	/**
	 * onShutdown is called whenever an intelligent sensor is shutdown and not longer available for SEFU/IFU
	 */
	@Override
	public void onShutdown(UaClient uaClient, long secondsTillShutdown, LocalizedText shutdownReason) {
		// Called when the server state changes to Shutdown
		// System.out.println(String.format("Server %s is not available.\nServer
		// shutdown in %d seconds.\nReason: %s\n",
		// uaClient.getUri(), secondsTillShutdown, shutdownReason.getText()));

		String uidISens = uaClient.getApplicationIdentity().getApplicationDescription().getProductUri();
		System.out.println("\n\nShutdown: " + uidISens);
		KB.removeIntelligenSensor(uidISens);
		removeClient(uaClient);
		this.runOrchestration();
	}

	// *******************************************************************************************************************
	/**
	 * onStateChanged is called whenever an intelligent sensor (OPC UA server) is detached from the network for example
	 * because of a wiring defect.
	 */
	@Override
	public void onStateChange(UaClient uaClient, ServerState oldState, ServerState newState) {
		// Called whenever the server state changes
		// System.out.println(String.format("ServerState changed from %s to %s
		// (%s)", oldState, newState, uaClient.getUri()));
		if (newState.equals(ServerState.Unknown))
			System.out.println(String.format("ServerStatusError: " + uaClient.getServerStatusError()));
		if (newState.equals(ServerState.CommunicationFault)) {
			String uidISens = uaClient.getApplicationIdentity().getApplicationDescription().getProductUri();
			System.out.println("\nDisconnected: " + uidISens);
			KB.removeIntelligenSensor(uidISens);
			this.runOrchestration();
		}
		if (newState.equals(ServerState.Running) && oldState.equals(ServerState.CommunicationFault)) {
			String uidISens = uaClient.getApplicationIdentity().getApplicationDescription().getProductUri();
			String[] parts = uaClient.getServerIdentity().getApplicationDescription().getApplicationUri().split(":");
			String ip = parts[1];
			System.out.println("\nReconnected: " + uidISens + "    IP: " + ip);

			try {
				this.readIntSensInfo(uaClient, ip, KB);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			this.runOrchestration();
		}
	}

	// *******************************************************************************************************************
	@Override
	public void onStatusChange(UaClient uaClient, ServerStatusDataType status) {
		// TODO Auto-generated method stub
	}
}
